#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.08.2019


'''
PEPT
====

Summary
-------
A Python library that unifies Positron Emission Particle Tracking (PEPT)
research, including tracking, simulation, data analysis and visualisation
tools.

Extended Summary
----------------
This Python library integrates all the tools necessary to perform research
using Positron Emission Particle Tracking (PEPT). The library includes
algorithms for the location, identification and tracking of particles, in
addition to tools for visualisation and analysis, and utilities allowing the
realistic simulation of PEPT data.

PEPT is a technique developed at the University of Birmingham [1]_ which allows
the non-invasive, three-dimensional tracking of one or more 'tracer' particles
through particulate, fluid or multiphase systems. The technique allows particle
or fluid motion to be tracked with sub-millimetre accuracy and sub-millisecond
temporal resolution and, due to its use of highly-penetrating 511keV gamma
rays, can be used to probe the internal dynamics of even large, dense,
optically opaque systems - making it ideal for industrial as well as
scientific applications.

PEPT is performed by radioactively labelling a particle with a positron-
emitting radioisotope such as fluorine-18 (18F) or gallium-66 (66Ga), and using
the back-to-back gamma rays produced by electron-positron annihilation events
in and around the tracer to triangulate its spatial position. Each detected
gamma ray represents a line of response (LoR) .

Subpackages Provided
--------------------
The `pept` package provides five base classes at its root - `PointData`,
`LineData`, `Pixels`, `Voxels` and `VoxelData` - acting as common data formats
that the rest of the package will use for its tracking, analysis and
visualisation algorithms. Thus any subroutine integrated in the `pept` package
can be used interchangeably with new PET / PEPT scanner geometries, data
formats or novel algorithms.

The rest of the package is grouped into subpackages and modules following the
hierarchy below:

::

    pept
    │
    Base classes imported into the package root:
    ├── PointData :     Base class encapsulating points.
    ├── LineData :      Base class encapsulating lines (LoRs) w/ one timestamp.
    ├── Pixels :        Base class managing pixels from a sample of lines.
    ├── Voxels :        Base class managing voxels from a sample of lines.
    ├── VoxelData :     Base class asynchronously managing multiple `Voxels`.
    │
    Subpackages:
    ├── base :                  Base classes (above).
    ├── cookbook :              Pre-made PEPT analysis scripts, or recipes.
    ├── diagnostics :           PET/PEPT scanner diagnostics.
    ├── processing :            PEPT-oriented post-processing algorithms.
    ├── scanners :              Transform other data formats into base classes.
    ├── simulation :            Simulate radioactively-labeled tracers.
    ├── tests :                 Package unit tests.
    ├── tracking :              Tracer identification and tracking algorithms.
    │   ├── birmingham_method : The original Birmingham Method [1].
    │   ├── fpi :               The Feature Point Identification algorithm [3].
    │   ├── peptml :            The PEPT-ML algorithm [2].
    │   └── trajectory_separation : Trajectory separation of tracked tracers.
    ├── utilities :             Utility functions such as fast CSV-readers.
    │   ├── cutpoints :         Compute cutpoints from LoRs; Cython functions.
    │   ├── misc :              Miscellaneous, I/O, data aggregation.
    │   ├── parallel :          Call arbitrary functions using multithreading.
    │   └── traverse :          Traverse voxels for LoRs; Cython functions.
    └── visualisation :         Visualisation algorithms for LoRs, points, etc.


Each subpackage has its own documentation containing module, class and function
hierarchies such as the one above.

Performance
-----------
Significant effort has been put into making the algorithms in this package as
fast as possible. The most compute-intensive parts have been implemented in
`C` / `Cython` and parallelised, where possible, using `joblib` and
`concurrent.futures.ThreadPoolExecutor`. For example, using the `peptml`
subpackage [2]_, analysing 1,000,000 LoRs on the author's machine (mid 2012
MacBook Pro) takes ~26 seconds.

Citing
------
If you used this codebase or any software making use of it in a scientific
publication, we ask you to cite the following paper:

    Nicuşan AL, Windows-Yule CR. Positron emission particle tracking using
    machine learning. Review of Scientific Instruments. 2020 Jan 1;91(1):013329
    https://doi.org/10.1063/1.5129251

Licensing
---------
The `pept` package is GNU v3.0 licensed.
Copyright (C) 2019-2021 the pept developers.

References
----------
Papers presenting PEPT algorithms included in this library: [1]_, [2]_, [3]_.

.. [1] Parker DJ, Broadbent CJ, Fowles P, Hawkesworth MR, McNeil P. Positron
   emission particle tracking-a technique for studying flow within engineering
   equipment. Nuclear Instruments and Methods in Physics Research Section A:
   Accelerators, Spectrometers, Detectors and Associated Equipment. 1993
   Mar 10;326(3):592-607.
.. [2] Nicuşan AL, Windows-Yule CR. Positron emission particle tracking using
   machine learning. Review of Scientific Instruments. 2020 Jan 1;91(1):013329.
.. [3] Wiggins C, Santos R, Ruggles A. A feature point identification method
   for positron emission particle tracking with multiple tracers. Nuclear
   Instruments and Methods in Physics Research Section A: Accelerators,
   Spectrometers, Detectors and Associated Equipment. 2017 Jan 21;843:22-8.

Examples
--------
You can download data samples from the UoB Positron Imaging Centre's
Repository (`link <https://bit.ly/pept-example-data-repo>`_). A
small but complete analysis script using the PEPT-ML algorithm is given below.

First import the `pept` package and read in lines of response (LoRs) from our
online repository. We'll use a sample of real data from an experiment conducted
at Birmingham - two tracers rotating at 42RPM.

.. code-block:: python

    import pept

    lors_raw = pept.utilities.read_csv(
        ("https://raw.githubusercontent.com/uob-positron-imaging-centres/"
         "example_data/master/sample_2p_42rpm.csv"),  # Concatenate long string
        skiprows = 16                                 # Skip file header
    )

Our data is in a format typical of the parallel screens PEPT detector from
Birmingham. Convert it to our package's scanner-agnostic line format; we'll
iterate through them one sample at a time, tracking the tracer locations:

.. code-block:: python

    lors = pept.scanners.ParallelScreens(
        lors_raw, screen_separation = 712, sample_size = 200
    )

Now it's time for some machine learning! Import the `peptml` package and
transform the LoRs into *cutpoints* that we'll use to find tracer locations.

.. code-block:: python

    from pept.tracking import peptml
    cutpoints = peptml.Cutpoints(lors, max_distance = 0.15)

Create a *clusterer* and use the cutpoints to find the centres of the tracers
as they move through our system:

.. code-block:: python

    clusterer = peptml.HDBSCANClusterer()
    centres = clusterer.fit(cutpoints)

This data looks alright (we'll plot everything in a minute), but we can do
better. We can iterate through the centres once more, clustering them again,
to get smooth, tight trajectories; for this, we'll set a small sample size and
large overlap:

.. code-block:: python

    centres.sample_size = 60
    centres.overlap = 59
    centres_2pass = clusterer.fit(centres)

One more step! The locations of the two tracers tracked above are intertwined;
we need to separate them into individual trajectories. We have a module for
just that:

.. code-block:: python

    import pept.tracking.trajectory_separation as tsp
    points_window = 10
    trajectory_cut_distance = 20
    trajectories = tsp.segregate_trajectories(
        centres_2pass, points_window, trajectory_cut_distance
    )

Finally, let's plot some LoRs, the tracer centres after one pass of clustering
and after the second pass of clustering, and the separated trajectories. Let's
use Plotly to produce some beautiful, interactive, 3D graphs:

.. code-block:: python

    from pept.visualisation import PlotlyGrapher
    subplot_titles = ["Lines of Response (LoRs)", "HDBSCAN Clustering",
                      "2-pass Clustering", "Separated Trajectories"]
    grapher = PlotlyGrapher(rows = 2, cols = 2, zlim = [0, 712],
                            subplot_titles = subplot_titles)

    grapher.add_lines(lors.lines[:400])             # Only the first 400 LoRs
    grapher.add_points(centres, col = 2)
    grapher.add_points(centres_2pass, row = 2)
    grapher.add_points(trajectories, row = 2, col = 2)

    grapher.show()

A more in-depth tutorial is available on
`Google Colab <https://bit.ly/pept-tutorial-1>`_.

'''


# Import base data structures
from    .base.line_data     import  LineData
from    .base.point_data    import  PointData
from    .base.pixel_data    import  Pixels
from    .base.voxel_data    import  Voxels
from    .base.voxel_data    import  VoxelData

# Import subpackages
from    .                   import  cookbook
from    .                   import  diagnostics
from    .                   import  processing
from    .                   import  scanners
from    .                   import  simulation
from    .                   import  tracking
from    .                   import  utilities
from    .                   import  visualisation

# Import package version
from    .__version__        import  __version__


__all__ = [
    'LineData',
    'PointData',
    'Pixels',
    'Voxels',
    'VoxelData',
    'cookbook',
    'diagnostics',
    'processing',
    'scanners',
    'simulation',
    'tracking',
    'utilities',
    'visualisation',
]


__author__ = ["The pept developers"]
__credits__ = [
    "Andrei Leonard Nicusan (University of Birmingham)",
    "Dr. Kit Windows-Yule (University of Birmingham)",
    "Dr. Sam Manger (University of Birmingham)",
    "Matthew Herald (University of Birmingham)",
    "Chris Jones (University of Birmingham)",
    "Prof. David Parker (University of Birmingham)",
    "Dr. Antoine Renaud (University of Edinburgh)",
    "Dr. Cody Wiggins (Virginia Commonwealth University",
]
__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : point_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.08.2019


import  pickle
import  time

import  numpy                   as      np

import  plotly.graph_objects    as      go
import  matplotlib.pyplot       as      plt

from    .iterable_samples       import  IterableSamples
import  pept


class PointData(IterableSamples):
    '''A class for general PEPT point-like data iteration, manipulation and
    visualisation.

    In the context of positron-based particle tracking, points are defined by a
    timestamp, 3D coordinates and any other extra information (such as
    trajectory label or some tracer signature). This class is used for the
    encapsulation of 3D points - be they tracer locations, cutpoints, etc. -,
    efficiently yielding samples of `points` of an adaptive `sample_size` and
    `overlap`.

    Much like a complement to `LineData`, `PointData` is an abstraction over
    point-like data that may be encountered in the context of PEPT (e.g.
    pre-tracked tracer locations), as once the raw points are transformed into
    the common `PointData` format, any tracking, analysis or visualisation
    algorithm in the `pept` package can be used interchangeably. Moreover, it
    provides a stable, user-friendly interface for iterating over points in
    *samples* - this can be useful for tracking algorithms, as some take a few
    points (a *sample*), produce an accurate tracer location, then move to the
    next sample of points, repeating the procedure. Using overlapping samples
    is also useful for improving the time resolution of the algorithms.

    This is the base class for point-like data; subroutines that accept and/or
    return `PointData` instances (or subclasses thereof) can be found
    throughout the `pept` package. If you'd like to create new algorithms based
    on them, you can check out the `pept.tracking.peptml.cutpoints` module as
    an example; the `Cutpoints` class receives a `LineData` instance,
    transforms the samples of LoRs into cutpoints, then initialises itself as a
    `PointData` subclass - thereby inheriting all its methods and attributes.

    Attributes
    ----------
    points : (N, M) numpy.ndarray
        An (N, M >= 4) numpy array that stores the points as time, followed by
        cartesian (3D) coordinates of the point, followed by any extra
        information. The data columns are then `[time, x, y, z, etc]`.

    sample_size : int
        An `int` that defines the number of lines that should be returned when
        iterating over `points`. The default is 0.

    overlap : int
        An `int` that defines the overlap between two consecutive samples that
        are returned when iterating over `points`. An overlap of 0 means
        consecutive samples, while an overlap of (`sample_size` - 1) means
        incrementing the samples by one. A negative overlap means skipping
        values between samples. It is required to be smaller than
        `sample_size`. The default is 0.

    number_of_points : int
        An `int` that corresponds to len(`points`), or the number of points
        stored by `points`.

    number_of_samples : int
        An `int` that corresponds to the number of samples that can be accessed
        from the class, taking the `overlap` into consideration.

    Methods
    -------
    sample(n)
        Get sample number n (indexed from 0).

    to_csv(filepath)
        Write `points` to a CSV file.

    save(filepath)
        Save a `PointData` instance as a binary `pickle` object.

    load(filepath)
        Load a saved / pickled `PointData` object from `filepath`.

    plot(sample_indices = ..., ax = None, alt_axes = False, colorbar_col = -1)
        Plot points from selected samples using matplotlib.

    points_trace(sample_indices = ..., size = 2, color = None, opacity = 0.8,\
                 colorbar = True, colorbar_col = -1, colorscale = "Magma",\
                 colorbar_title = None)
        Get a Plotly trace for all points in selected samples, with possible
        color-coding.

    copy()
        Create a deep copy of an instance of this class, including a new inner
        numpy array `points`.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size`. Overlap is required to be smaller than
        `sample_size`, unless `sample_size` is 0. Note that it can also be
        negative.

    Notes
    -----
    This class saves `points` as a **contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the number of rows or columns after instantiating the class.

    Examples
    --------
    Initialise a `PointData` instance containing 10 points with a `sample_size`
    of 3.

    .. ipython::

        In [1]: import numpy as np

        In [2]: import pept

        In [3]: points_raw = np.arange(40).reshape(10, 4)

        In [4]: print(points_raw)
        [[ 0  1  2  3]
         [ 4  5  6  7]
         [ 8  9 10 11]
         [12 13 14 15]
         [16 17 18 19]
         [20 21 22 23]
         [24 25 26 27]
         [28 29 30 31]
         [32 33 34 35]
         [36 37 38 39]]

        In [5]: point_data = pept.PointData(points_raw, sample_size = 3)

        In [6]: print(point_data)
        number_of_points =  10
        sample_size =       3
        overlap =           0
        number_of_samples = 3
        points =
        [[ 0.  1.  2.  3.]
         [ 4.  5.  6.  7.]
         [ 8.  9. 10. 11.]
         [12. 13. 14. 15.]
         [16. 17. 18. 19.]
         [20. 21. 22. 23.]
         [24. 25. 26. 27.]
         [28. 29. 30. 31.]
         [32. 33. 34. 35.]
         [36. 37. 38. 39.]]

    Access samples using subscript notation. Notice how the samples are
    consecutive, as `overlap` is 0 by default.

    .. ipython::

        In [7]: point_data[0]
        Out[7]:
        array([[ 0.,  1.,  2.,  3.],
               [ 4.,  5.,  6.,  7.],
               [ 8.,  9., 10., 11.]])

        In [8]: point_data[1]
        Out[8]:
        array([[12., 13., 14., 15.],
               [16., 17., 18., 19.],
               [20., 21., 22., 23.]])

    Now set an overlap of 2; notice how the number of samples changes:

    .. ipython::

        In [9]: len(point_data)         # Number of samples
        Out[9]: 3

        In [10]: point_data.overlap = 2

        In [11]: len(point_data)
        Out[11]: 8

    Notice how rows are repeated from one sample to the next when accessing
    them, because `overlap` is now 2:

    .. ipython::

        In [12]: point_data[0]
        Out[12]:
        array([[ 0.,  1.,  2.,  3.],
               [ 4.,  5.,  6.,  7.],
               [ 8.,  9., 10., 11.]])

        In [13]: point_data[1]
        Out[13]:
        array([[ 4.,  5.,  6.,  7.],
               [ 8.,  9., 10., 11.],
               [12., 13., 14., 15.]])

    Now change `sample_size` to 5 and notice again how the number of samples
    changes:

    .. ipython::

        In [14]: len(point_data)
        Out[14]: 8

        In [15]: point_data.sample_size = 5

        In [16]: len(point_data)
        Out[16]: 2

        In [17]: point_data[0]
        Out[17]:
        array([[ 0.,  1.,  2.,  3.],
               [ 4.,  5.,  6.,  7.],
               [ 8.,  9., 10., 11.],
               [12., 13., 14., 15.],
               [16., 17., 18., 19.]])

        In [18]: point_data[1]
        Out[18]:
        array([[12., 13., 14., 15.],
               [16., 17., 18., 19.],
               [20., 21., 22., 23.],
               [24., 25., 26., 27.],
               [28., 29., 30., 31.]])

    Notice how the samples do not cover the whole input `points_raw` array, as
    the last lines are omitted - think of the `sample_size` and `overlap`. They
    are still inside the inner `points` attribute of `point_data` though:

    .. ipython::

        In [19]: point_data.points
        Out[19]:
        array([[ 0.,  1.,  2.,  3.],
               [ 4.,  5.,  6.,  7.],
               [ 8.,  9., 10., 11.],
               [12., 13., 14., 15.],
               [16., 17., 18., 19.],
               [20., 21., 22., 23.],
               [24., 25., 26., 27.],
               [28., 29., 30., 31.],
               [32., 33., 34., 35.],
               [36., 37., 38., 39.]])

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    pept.tracking.peptml.Cutpoints : Compute cutpoints from `pept.LineData`.
    '''

    def __init__(
        self,
        points,
        sample_size = 0,
        overlap = 0,
        verbose = False
    ):
        '''`PointData` class constructor.

        Parameters
        ----------
        points : (N, M) numpy.ndarray
            An (N, M >= 4) numpy array that stores points (or any generic 2D
            set of data). It expects that the first column is time, followed by
            cartesian (3D) coordinates of points, followed by any extra
            information the user needs. The data columns are then
            `[time, x, y, z, etc]`.

        sample_size : int, default 0
            An `int`` that defines the number of points that should be returned
            when iterating over `points`. A `sample_size` of 0 yields all the
            data as one single sample.

        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `points`. An overlap of 0
            means consecutive samples, while an overlap of (`sample_size` - 1)
            implies incrementing the samples by one. A negative overlap means
            skipping values between samples. An error is raised if `overlap` is
            larger than or equal to `sample_size`.

        verbose : bool, default False
            An option that enables printing the time taken for the
            initialisation of an instance of the class. Useful when reading
            large files (10gb files for PEPT data is not unheard of).

        Raises
        ------
        ValueError
            If `line_data` does not have (N, M) shape, where M >= 4.
        '''

        if verbose:
            start = time.time()

        # If `points` is not C-contiguous, create a C-contiguous copy.
        self._points = np.asarray(points, order = 'C', dtype = float)

        # Check that points has at least 4 columns.
        if self._points.ndim != 2 or self._points.shape[1] < 4:
            raise ValueError((
                "\n[ERROR]: `points` should have dimensions (M, N), where "
                f"N >= 4. Received {self._points.shape}.\n"
            ))

        self._number_of_points = len(self._points)

        # Call the IterableSamples constructor to make the class iterable in
        # samples with overlap.
        IterableSamples.__init__(self, sample_size, overlap)

        if verbose:
            end = time.time()
            print(f"Initialised PointData class in {end - start} s.")


    @property
    def points(self):
        return self._points


    @property
    def data_samples(self):
        # Implemented property for the `IterableSamples` parent class. See its
        # documentation for more details.
        return self._points


    @property
    def data_length(self):
        # Implemented property for the IterableSamples parent class. See its
        # documentation for more details.
        return self._number_of_points


    @property
    def number_of_points(self):
        return self._number_of_points


    def to_csv(self, filepath):
        '''Write `points` to a CSV file.

        Write all points (and any extra data) stored in the class to a CSV
        file.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.
        '''

        np.savetxt(filepath, self._points)


    def save(self, filepath):
        '''Save a `PointData` instance as a binary `pickle` object.

        Saves the full object state, including the inner `.points` NumPy array,
        `sample_size`, etc. in a fast, portable binary format. Load back the
        object using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `PointData` instance, then load it back:

        >>> points = pept.PointData([[1, 2, 3, 4]])
        >>> points.save("points.pickle")

        >>> points_reloaded = pept.PointData.load("points.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `PointData` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.PointData
            The loaded `pept.PointData` instance.

        Examples
        --------
        Save a `PointData` instance, then load it back:

        >>> points = pept.PointData([[1, 2, 3, 4]])
        >>> points.save("points.pickle")

        >>> points_reloaded = pept.PointData.load("points.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def plot(
        self,
        sample_indices = ...,
        ax = None,
        alt_axes = False,
        colorbar_col = -1,
    ):
        '''Plot points from selected samples using matplotlib.

        Returns matplotlib figure and axes objects containing all points
        included in the samples selected by `sample_indices`.
        `sample_indices` may be a single sample index (e.g. 0), an iterable
        of indices (e.g. [1,5,6]), or an Ellipsis (`...`) for all samples.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of points. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all points).

        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        alt_axes : bool, default False
            If `True`, plot using the alternative PEPT-style axes convention:
            z is horizontal, y points upwards. Because Matplotlib cannot swap
            axes, this is achieved by swapping the parameters in the plotting
            call (i.e. `plt.plot(x, y, z)` -> `plt.plot(z, x, y)`).

        colorbar_col : int, default -1
            The column in the data samples that will be used to color the
            points. The default is -1 (the last column).

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Notes
        -----
        Plotting all points is very computationally-expensive for matplotlib.
        It is recommended to only plot a couple of samples at a time, or use
        the faster `pept.visualisation.PlotlyGrapher`.

        Examples
        --------
        Plot the points from sample 1 in a `PointData` instance:

        >>> point_data = pept.PointData(...)
        >>> fig, ax = point_data.plot(1)
        >>> fig.show()

        Plot the points from samples 0, 1 and 2:

        >>> fig, ax = point_data.plot([0, 1, 2])
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection = '3d')
        else:
            fig = plt.gcf()

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        if sample_indices[0] == Ellipsis:
            x = self._points[:, 1],
            y = self._points[:, 2],
            z = self._points[:, 3],
            color = self._points[:, colorbar_col]
        else:
            x, y, z, color = [], [], [], []
            for n in sample_indices:
                sample = self[n]

                x.extend(sample[:, 1])
                y.extend(sample[:, 2])
                z.extend(sample[:, 3])

                color.extend(sample[:, colorbar_col])

        color = np.array(color)

        # Scatter x, y, z, [color]
        cmap = plt.cm.magma
        color_array = cmap(color / color.max())

        if alt_axes:
            ax.scatter(z, x, y, c = color_array)

            ax.set_xlabel("z (mm)")
            ax.set_ylabel("x (mm)")
            ax.set_zlabel("y (mm)")

        else:
            ax.scatter(x, y, z, c = color_array)

            ax.set_xlabel("x (mm)")
            ax.set_ylabel("y (mm)")
            ax.set_zlabel("z (mm)")

        return fig, ax


    def points_trace(
        self,
        sample_indices = ...,
        size = 2,
        color = None,
        opacity = 0.8,
        colorbar = True,
        colorbar_col = -1,
        colorscale = "Magma",
        colorbar_title = None
    ):
        '''Get a Plotly trace for all points in selected samples, with possible
        color-coding.

        Returns a `plotly.graph_objects.Scatter3d` trace containing all points
        included in in the samples selected by `sample_indices`.
        `sample_indices` may be a single sample index (e.g. 0), an iterable
        of indices (e.g. [1,5,6]), or an Ellipsis (`...`) for all samples.

        Can then be passed to the `plotly.graph_objects.figure.add_trace`
        function or a `PlotlyGrapher` instance using the `add_trace` method.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of points. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all points).

        size : float, default 2
            The marker size of the points.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation.

        opacity : float, default 0.8
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the data in the sample column
            `colorbar_col`. Is overridden if `color` is set.

        colorbar_col : int, default -1
            The column in the data samples that will be used to color the
            points. Only has an effect if `colorbar` is set to True. The
            default is -1 (the last column).

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the `colorbar_col` column in the
            input data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above.

        Returns
        -------
        plotly.graph_objs.Scatter3d
            A Plotly trace of the points.

        Examples
        --------
        Use `PlotlyGrapher` (a user-friendly wrapper around the `plotly`
        library for PEPT-oriented data) to plot the points from sample 1 in a
        `PointData` instance:

        >>> point_data = pept.PointData(...)
        >>> grapher = pept.visualisation.PlotlyGrapher()
        >>> trace = point_data.points_trace(1)
        >>> grapher.add_trace(trace)
        >>> grapher.show()

        Use `plotly.graph_objs` to plot the lines from samples 0, 1 and 2:

        >>> import plotly.graph_objs as go
        >>> fig = go.Figure()
        >>> fig.add_trace(point_data.points_trace([0, 1, 2]))
        >>> fig.show()

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        coords_x = []
        coords_y = []
        coords_z = []

        marker = dict(
            size = size,
            color = color,
            opacity = opacity
        )

        if colorbar:
            if color is None:
                marker['color'] = []

            marker.update(colorscale = colorscale)
            if colorbar_title is not None:
                marker.update(colorbar = dict(title = colorbar_title))

        # If an Ellipsis was received, include all points
        if sample_indices[0] is Ellipsis:
            coords_x = self.points[:, 1]
            coords_y = self.points[:, 2]
            coords_z = self.points[:, 3]

            if colorbar and color is None:
                marker['color'] = self.points[:, colorbar_col]
        else:
            # For each selected sample include all the needed coordinates
            for n in sample_indices:
                sample = self[n]

                coords_x.extend(sample[:, 1])
                coords_y.extend(sample[:, 2])
                coords_z.extend(sample[:, 3])

                if colorbar and color is None:
                    marker['color'].extend(sample[:, colorbar_col])

        trace = go.Scatter3d(
            x = coords_x,
            y = coords_y,
            z = coords_z,
            mode = "markers",
            marker = marker
        )

        return trace


    def copy(self):
        '''Create a deep copy of an instance of this class, including a new
        inner numpy array `points`.

        Returns
        -------
        pept.PointData
            A new instance of the `pept.PointData` class with the same
            attributes as this instance, deep-copied.
        '''

        return pept.PointData(
            self._points.copy(order = "C"),
            sample_size = self._sample_size,
            overlap = self._overlap,
            verbose = False
        )


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"number_of_points =  {self.number_of_points}\n\n"
            f"sample_size =       {self._sample_size}\n"
            f"overlap =           {self._overlap}\n"
            f"number_of_samples = {self.number_of_samples}\n\n"
            f"points = \n{self._points}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL

        docstr = (
            "Class instance that inherits from `pept.PointData`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}\n\n"
            "Particular Cases\n----------------\n"
            " > If sample_size == 0, all `points` are returned as a "
            "single sample.\n"
            " > If overlap >= sample_size, an error is raised.\n"
            " > If overlap < 0, points are skipped between samples."
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2020 Andrei Leonard Nicusan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : adac_forte.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 01.04.2021


import  os
import  time
import  textwrap

from    pept            import  LineData

from    .extensions     import  convert_adac_forte




class ADACForte(LineData):
    '''A subclass of `LineData` that initialises PEPT lines of response (LoRs)
    from the ADAC Forte parallel screen detector list mode *binary* format.

    Inherits all properties and methods from the general `LineData` class, but
    reads in LoRs from ADAC binary files (usual extension ".da01").

    Attributes
    ----------
    sample_size, overlap, number_of_lines, etc.: inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Methods
    -------
    to_csv, lines_trace, etc. : inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Examples
    --------
    Initialise a `ParallelScreens` array for three LoRs on a parallel screens
    PEPT scanner (i.e. each line is defined by **two** points each) with a
    head separation of 500 mm:

    >>> lors = pept.scanners.ADACForte("binary_data_adac.da01")
    >>> Initialising the PEPT data took 0.00038814544677734375 seconds.

    >>> lors
    >>> Class instance that inherits from `pept.LineData`.
    >>> Type:
    >>> <class 'pept.scanners.parallel_screens.adac_forte.ADACForte'>
    >>>
    >>> Attributes
    >>> ----------
    >>> number_of_lines =   3
    >>>
    >>> sample_size =       0
    >>> overlap =           0
    >>> number_of_samples = 1
    >>>
    >>> lines =
    >>> [[  2. 100. 150.   0. 200. 250. 500.]
    >>>  [  4. 350. 250.   0. 100. 150. 500.]
    >>>  [  6. 450. 350.   0. 250. 200. 500.]]
    >>>
    >>> Particular Cases
    >>> ----------------
    >>> > If sample_size == 0, all `lines` are returned as a single sample.
    >>> > If overlap >= sample_size, an error is raised.
    >>> > If overlap < 0, lines are skipped between samples.

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    def __init__(
        self,
        filepath,
        sample_size = 0,
        overlap = 0,
        verbose = True
    ):
        '''ParallelScreens class constructor.

        Parameters
        ----------
        filepath_or_array : str
            The path to a ADAC Forte-generated binary file from which the LoRs
            will be read into the `LineData` format.

        sample_size : int, default 0
            An `int` that defines the number of lines that should be returned
            when iterating over `lines`. A `sample_size` of 0 yields all the
            data as one single sample. A good starting value would be 200 times
            the maximum number of tracers that would be tracked.

        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `lines`. An overlap of 0
            implies consecutive samples, while an overlap of
            (`sample_size` - 1) means incrementing the samples by one. A
            negative overlap means skipping values between samples. An error is
            raised if `overlap` is larger than or equal to `sample_size`.

        verbose : bool, default True
            An option that enables printing the time taken for the
            initialisation of an instance of the class. Useful when reading
            large files (10gb files for PEPT data is not unheard of).

        Raises
        ------
        FileNotFoundError
            If the input `filepath` does not exist.

        ValueError
            If `overlap` >= `sample_size`. Overlap has to be smaller than
            `sample_size`. Note that it can also be negative.
        '''

        if verbose:
            start = time.time()

        if not os.path.isfile(filepath):
            raise FileNotFoundError(textwrap.fill((
                f"The input file path {filepath} does not exist!"
            )))

        lines = convert_adac_forte(filepath)

        # Call the constructor of the superclass `LineData` to initialise all
        # the inner parameters of the class (_index, etc.)
        LineData.__init__(
            self,
            lines,
            sample_size = sample_size,
            overlap = overlap,
            verbose = False
        )

        if verbose:
            end = time.time()
            print(f"Initialising the PEPT data took {end - start} seconds.\n")



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


'''The `birmingham_method` package provides an efficient, optionally-parallel
implementation of the well-known Birmingham method for single-tracer tracking.

Summary
-------
A typical workflow for using the `birmingham_method` package would be:

1. Read the LoRs into a `pept.LineData` class instance and set the
   `sample_size` and `overlap` appropriately.
2. Instantiate a `pept.tracking.birmingham_method.BirminghamMethod` class and
   transform the LoRs into tracer locations using the `fit` method.

Extended Summary
----------------
For a given "sample" of LoRs, the Birmingham method minimises the distance
between all of the LoRs, rejecting a fraction of lines that lie furthest away
from the calculated distance. The process is repeated iteratively until a
specified fraction ("fopt") of the original subset of LORs remains.

The Birmingham method has been used extensively for well over 30 years at the
University of Birmingham to track radioactively-labelled tracers in a variety
of industrial and scientific systems [1]_.

Modules Provided
----------------

::

    pept.tracking.birmingham_method
    │
    Classes imported into the subpackage root:
    └── BirminghamMethod :  Transform samples of LoRs into tracer locations.

References
----------
.. [1] Parker DJ, Broadbent CJ, Fowles P, Hawkesworth MR, McNeil P. Positron
   emission particle tracking-a technique for studying flow within engineering
   equipment. Nuclear Instruments and Methods in Physics Research Section A:
   Accelerators, Spectrometers, Detectors and Associated Equipment. 1993 Mar
   10;326(3):592-607.
'''


from    .birmingham_method    import BirminghamMethod


__all__ = [
    'BirminghamMethod',
]


__license__ = "GNU v3.0"
__maintainer__ = "Sam Manger"
__email__ = "s.manger@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 31.01.2020


'''PEPT base classes.

Subpackages Provided
--------------------
These classes are imported into the ``pept`` package root, so they are
available as ``pept.PointData``, ``pept.LineData``, etc., without going through
``base`` first.

::

    pept.base
    ├── PointData :     Encapsulate points for iteration and visualisation.
    ├── LineData :      Encapsulate lines (LoRs) with a single timestamp.
    ├── Pixels :        Encapsulate pixels for a single sample of lines.
    ├── Voxels :        Encapsulate voxels for a single sample of lines.
    └── VoxelData :     Encapsulate voxels for line traversal and manipulation.

'''


from    .line_data      import  LineData
from    .point_data     import  PointData
from    .voxel_data     import  VoxelData
from    .voxel_data     import  Voxels
from    .pixel_data     import  Pixels


__all__ = [
    'LineData',
    'PointData',
    'Pixels',
    'Voxels',
    'VoxelData',
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you should cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : iterable_samples.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 09.04.2020


from abc import ABC, abstractmethod


class IterableSamples(ABC):
    '''An abstract class for iterating through an array (or array-like) in
    samples with potential overlap.

    A class that inherits from ``IterableSamples`` must implement two
    properties:

    ``data_samples``
        The underlying data (e.g. a numpy array) that supports slicing.

    ``data_length``
        The number of items in ``data_samples`` (e.g. for a 2D numpy array,
        this is the number of rows).

    Given an implementor's properties ``data_samples`` (e.g. a numpy array) and
    ``data_length``, this class can yield samples of the data of an adaptive
    ``sample_size`` and ``overlap``, without requiring additional storage.

    The underlying data can be accessed using both indexing (``data[0]``) and
    iteration (``for sample in data: ...``).

    Particular cases:
        1. If sample_size == 0, all data_samples is returned as one single
           sample.
        2. If overlap >= sample_size, an error is raised.
        3. If overlap < 0, lines are skipped between samples.

    Attributes
    ----------
    data_samples : iterable that supports slicing
        An iterable (e.g. numpy array) that supports slicing syntax (data[5:7])
        storing the data that will be iterated over in samples. Must be
        implemented by a subclass!

    data_length : int
        The number of elements in `data_samples`. For a numpy.ndarray, that
        corresponds to len(`line_data`). Must be implemented by a subclass!

    sample_size : int
        An `int`` that defines the number of items that should be returned in
        a single sample when iterating over `data_samples`. A `sample_size` of
        0 yields all the data as one single sample.

    overlap : int, optional
        An `int` that defines the overlap between two consecutive samples that
        are returned when iterating over `data_samples`. An overlap of 0
        implies consecutive samples, while an overlap of (`sample_size` - 1)
        means incrementing the samples by one. A negative overlap implies
        skipping values between samples. An error is raised if `overlap` is
        larger than or equal to `sample_size`.

    number_of_samples : int
        An `int` that corresponds to the number of samples that can be
        accessed from the class. It takes `overlap` into consideration.

    Methods
    -------
    sample(n)
        Get sample number n (indexed from 0).

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size` unless `sample_size` is 0. Overlap
        must be smaller than `sample_size`. Note that it can also be negative.

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.

    '''

    def __init__(self, sample_size, overlap):
        '''`IterableSamples` class constructor.

        Parameters
        ----------
        sample_size : int
            An `int`` that defines the number of items that should be returned
            in a single sample when iterating over `data_samples`. A
            `sample_size` of 0 yields all the data as one single sample.

        overlap : int, optional
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `data_samples`. An overlap of
            0 implies consecutive samples, while an overlap of
            (`sample_size` - 1) means incrementing the samples by one. A
            negative overlap implies skipping values between samples. An error
            is raised if `overlap` is larger than or equal to `sample_size`.
        '''

        sample_size = int(sample_size)
        overlap = int(overlap)

        if sample_size < 0:
            raise ValueError((
                f"\n[ERROR]: sample_size = {sample_size} must be positive "
                "(>= 0).\n"
            ))

        if sample_size != 0 and overlap >= sample_size:
            raise ValueError((
                f"\n[ERROR]: overlap = {overlap} must be smaller than "
                f"sample_size = {sample_size}.\n"
            ))

        self._index = 0
        self._sample_size = sample_size
        self._overlap = overlap


    @property
    @abstractmethod
    def data_samples(self):
        pass


    @property
    @abstractmethod
    def data_length(self):
        pass



    @property
    def sample_size(self):
        return self._sample_size


    @sample_size.setter
    def sample_size(self, sample_size):
        sample_size = int(sample_size)

        if sample_size < 0:
            raise ValueError((
                f"\n[ERROR]: sample_size = {sample_size} must be positive "
                "(>= 0). \n"
            ))

        if sample_size != 0 and self._overlap >= sample_size:
            raise ValueError((
                f"\n[ERROR]: overlap = {self._overlap} must be smaller than "
                f"sample_size = {sample_size}.\n"
            ))

        self._index = 0
        self._sample_size = sample_size


    @property
    def overlap(self):
        return self._overlap


    @overlap.setter
    def overlap(self, overlap):
        overlap = int(overlap)

        if self._sample_size != 0 and overlap >= self._sample_size:
            raise ValueError((
                f"\n[ERROR]: overlap = {overlap} must be smaller than "
                f"sample_size = {self._sample_size}.\n"
            ))

        self._index = 0
        self._overlap = overlap


    @property
    def number_of_samples(self):
        # If self.sample_size == 0, all data is returned as a single sample.
        if self._sample_size == 0:
            return 1

        # If self.sample_size != 0, check that there is at least one sample.
        if self.data_length >= self._sample_size:
            return (self.data_length - self._sample_size) // \
                (self.sample_size - self.overlap) + 1
        else:
            return 0


    def sample(self, n):
        '''Get sample number n (indexed from 0, i.e. `n >= 0`)

        Returns the items from `data_samples` included in sample number `n`.
        Samples are numbered starting from 0.

        Parameters
        ----------
        n : int
            The number of the sample required. Note that 1 <= n <=
            number_of_samples.

        Returns
        -------
        items from data_samples
            The items returned from `point_data` using slicing, included in
            sample `n`.

        Raises
        ------
        IndexError
            If `sample_size == 0`, all data is returned as one single sample.
            Raised if `n` is not 1.

        IndexError
            If `n > number_of_samples` or `n <= 0`.
        '''

        if self._sample_size == 0:
            if n == 0:
                return self.data_samples
            else:
                raise IndexError((
                    "\n[ERROR]: Tried to access a non-existent sample "
                    f"(samples indexed from 0): asked for sample {n}, when "
                    "there is only 1 sample (because `sample_size` == 0).\n"
                ))
        elif (n >= self.number_of_samples) or n < 0:
            raise IndexError((
                "\n[ERROR]: Tried to access a non-existent sample (samples "
                f"are indexed from 0): asked for sample {n}, when there are "
                f"{self.number_of_samples} samples.\n"
            ))

        start_index = n * (self._sample_size - self._overlap)
        return self.data_samples[start_index:(start_index + self._sample_size)]


    def __len__(self):
        # Defined so that len(class_instance) returns the number of samples.
        return self.number_of_samples


    def __getitem__(self, key):
        # Defined so that samples can be accessed as class_instance[0]
        while key < 0:
            key += self.number_of_samples

        return self.sample(key)


    def __iter__(self):
        # Defined so the class can be iterated as
        # `for sample in class_instance: ...`
        return self


    def __next__(self):
        # sample_size = 0 => return all data
        if self._sample_size == 0:
            self._sample_size = -1
            return self.data_samples
        # Use -1 as a flag
        if self._sample_size == -1:
            self._sample_size = 0
            raise StopIteration

        # sample_size > 0 => return slices
        if self._index != 0:
            self._index = self._index + self._sample_size - self._overlap
        else:
            self._index = self._index + self._sample_size

        if self._index > self.data_length:
            self._index = 0
            raise StopIteration

        return self.data_samples[(self._index - self._sample_size):self._index]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : parallel_screens.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.08.2019


import  time
import  numpy           as      np

from    pept            import  LineData
from    pept.utilities  import  read_csv




class ParallelScreens(LineData):
    '''A subclass of `LineData` that initialises PEPT data for parallel screens
    PET/PEPT detectors from an input data file or array.

    Provides the same functionality as the `LineData` class while initialising
    `lines` from  **PEPT scanners with two parallel screens**. That is,
    each LoR is defined by two 2D points on two screens separated by a given
    distance.

    **The expected data columns in the file are `[time, x1, y1, x2, y2]`**.
    This is automatically transformed into the standard `lines` format with
    columns being `[time, x1, y1, z1, x2, y2, z2]`, where `z1 = 0` and
    `z2 = screen_separation`.

    `ParallelScreens` can be initialised with a predefined numpy array of LoRs
    or read data from a `.csv` or `.a0n` file or equivalent. The attributes
    and methods are the same as for `pept.LineData`.

    Attributes
    ----------
    sample_size, overlap, number_of_lines, etc.: inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Methods
    -------
    to_csv, lines_trace, etc. : inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Examples
    --------
    Initialise a `ParallelScreens` array for three LoRs on a parallel screens
    PEPT scanner (i.e. each line is defined by **two** points each) with a
    head separation of 500 mm:

    >>> lors_raw = np.array([
    >>>     [2, 100, 150, 200, 250],
    >>>     [4, 350, 250, 100, 150],
    >>>     [6, 450, 350, 250, 200]
    >>> ])

    >>> screen_separation = 500
    >>> lors = pept.scanners.ParallelScreens(lors_raw, screen_separation)
    >>> Initialising the PEPT data took 0.00038814544677734375 seconds.

    >>> lors
    >>> Class instance that inherits from `pept.LineData`.
    >>> Type:
    >>> <class 'pept.scanners.parallel_screens.parallel_screens.ParallelScreens'>
    >>>
    >>> Attributes
    >>> ----------
    >>> number_of_lines =   3
    >>>
    >>> sample_size =       0
    >>> overlap =           0
    >>> number_of_samples = 1
    >>>
    >>> lines =
    >>> [[  2. 100. 150.   0. 200. 250. 500.]
    >>>  [  4. 350. 250.   0. 100. 150. 500.]
    >>>  [  6. 450. 350.   0. 250. 200. 500.]]
    >>>
    >>> Particular Cases
    >>> ----------------
    >>> > If sample_size == 0, all `lines` are returned as a single sample.
    >>> > If overlap >= sample_size, an error is raised.
    >>> > If overlap < 0, lines are skipped between samples.

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    def __init__(
        self,
        filepath_or_array,
        screen_separation,
        sample_size = 0,
        overlap = 0,
        skiprows = None,
        nrows = None,
        verbose = True
    ):
        '''ParallelScreens class constructor.

        Parameters
        ----------
        filepath_or_array : [str, pathlib.Path, IO] or numpy.ndarray (N, 5)
            A path to a file to be read from or an array for initialisation. A
            path is a string with the (absolute or relative) path to the data
            file or a URL from which the PEPT data will be read. It should
            include the full file name, along with its extension (.csv, .a01,
            etc.).
        screen_separation : float
            The separation (in *mm*) between the two PEPT screens corresponding
            to the `z` coordinate of the second point defining each line. The
            attribute `lines`, with columns
            `[time, x1, y1, z1, x2, y2, z2]`, will have `z1 = 0` and
            `z2 = screen_separation`.
        sample_size : int, default 0
            An `int` that defines the number of lines that should be returned
            when iterating over `lines`. A `sample_size` of 0 yields all the
            data as one single sample. A good starting value would be 200 times
            the maximum number of tracers that would be tracked.
        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `lines`. An overlap of 0
            implies consecutive samples, while an overlap of
            (`sample_size` - 1) means incrementing the samples by one. A
            negative overlap means skipping values between samples. An error is
            raised if `overlap` is larger than or equal to `sample_size`.
        skiprows : int, default 0
            The number of rows to skip from the beginning of the data file.
            Useful when the data file includes a header of text that should be
            skipped.
        max_rows : int, optional
            The maximum number of rows that will be read from the data file.
        verbose : bool, default True
            An option that enables printing the time taken for the
            initialisation of an instance of the class. Useful when reading
            large files (10gb files for PEPT data is not unheard of).

        Raises
        ------
        ValueError
            If `overlap` >= `sample_size`. Overlap has to be smaller than
            `sample_size`. Note that it can also be negative.
        ValueError
            If the data file does not have the (N, M >= 5) shape.
        '''

        if verbose:
            start = time.time()

        # Check wheter input is a valid `pandas.read_csv` filepath or a numpy
        # array in an "Easier To Ask Forgiveness Than Permission" way.
        try:
            # Try to read the LoR data from `filepath_or_array`.
            # Check if an error is raised when reading file lines using
            lines = read_csv(
                filepath_or_array,
                skiprows = skiprows,
                nrows = nrows
            )
        except ValueError:
            # Seems like it is an array!
            lines = np.asarray(
                filepath_or_array,
                order = "C",
                dtype = float
            )

        # lines cols: [time, X1, Y1, X2, Y2]
        # Verify that lines has shape (N, M >= 5)
        if lines.ndim != 2 or lines.shape[1] < 5:
            raise ValueError((
                "\n[ERROR]: lines should have dimensions (N, M) where "
                f"M >= 5. Received {lines.shape}.\n"
            ))

        # Add Z1 and Z2 columns => [time, X1, Y1, Z1, X2, Y2, Z2]
        # Z1 = 0
        lines = np.insert(lines, 3, 0.0, axis = 1)

        # Z2 = `separation`
        lines = np.insert(lines, 6, screen_separation, axis = 1)

        # Call the constructor of the superclass `LineData` to initialise all
        # the inner parameters of the class (_index, etc.)
        LineData.__init__(
            self,
            lines,
            sample_size = sample_size,
            overlap = overlap,
            verbose = False
        )

        if verbose:
            end = time.time()
            print(f"Initialising the PEPT data took {end - start} seconds.\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 14.01.2020


'''PEPT-oriented utility functions.

Summary
-------
The utility functions include low-level optimised Cython functions (e.g.
`find_cutpoints`) that are of common interest across the `pept` package, as
well as I/O functions, parallel maps and pixel/voxel traversal algorithms.

Even though the functions are grouped in directories (subpackages) and files
(modules), unlike the rest of the package, they are all imported into the
`pept.utilities` root, so that their import paths are not too long.

Subpackages Provided
--------------------

::

    pept.utilities
    │
    Functions imported into the subpackage root:
    ├── find_cutpoints :            Find cutpoints from an array of LoRs.
    ├── find_minpoints :            Find minpoints from an array of LoRs.
    ├── group_by_column :           Group array into a list by column.
    ├── number_of_lines :           Find the number of lines in a file.
    ├── read_csv :                  Fast CSV file reader.
    ├── read_csv_chunks :           Read CSV file in chunks using a generator.
    ├── parallel_map_file :         Map a function to chunks of a file.
    ├── traverse2d :                Traverse pixels.
    ├── traverse3d :                Traverse voxels.
    │
    Classes imported into the subpackage root:
    ├── ChunkReader :               Lazily read/access chunks from a CSV file.
    │
    Subpackages
    ├── cutpoints
    ├── misc
    ├── parallel
    └── traverse

'''


from    .cutpoints  import  *
from    .traverse   import  *
from    .parallel   import  *
from    .misc       import  *


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019 Andrei Leonard Nicusan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : modular_camera.py
# License: License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


import  os
import  time
import  textwrap

import  numpy                       as      np

from    pept                        import  LineData
from    .extensions.get_pept_event  import  get_pept_LOR


class ModularCamera(LineData):
    '''A subclass of `LineData` that reads PEPT data from the modular camera
    DAQ.

    Provides the same functionality as the `LineData` class while
    initialising `lines` from a given file. This is a helper class
    for **PEPT using the modular camera**.

    Can read data from a `.da_1` file or equivalent. The file must contain
    the standard datawords from the modular camera output. This will then
    be automatically transformed into the standard `_lines` format
    with every row being `[time, x1, y1, z1, x2, y2, z2]`, where the geometry
    is derived from the C-extension. The current useable geometry is a square
    layout with 4 stacks for 4 modules, separated by 250 mm.

    Attributes
    ----------
    sample_size, overlap, number_of_lines, etc.: inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Methods
    -------
    to_csv, lines_trace, etc. : inherited from `pept.LineData`
        All attributes and methods from the parent class `pept.LineData` are
        available after instantiation. Check its documentation for more
        information.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size`. Overlap has to be smaller than
        `sample_size`. Note that it can also be negative.
    ValueError
        If the data file does not have (N, 7) shape.

    Notes
    -----
    The class saves `lines` as a **contiguous** numpy array for
    efficient access in C functions. It should not be changed after
    instantiating the class.
    '''

    def __init__(
        self,
        data_file,
        sample_size = 200,
        overlap = 0,
        filtered = [],
        verbose = True
    ):
        '''`ModularCamera` class constructor.

        Parameters
        ----------
        data_file : str
            A string with the (absolute or relative) path to the data file
            from which the PEPT data will be read. It should include the
            full file name, along with the extension (.da_1)
        sample_size : int, optional
            An `int`` that defines the number of lines that should be
            returned when iterating over `_lines`. A `sample_size` of 0
            yields all the data as one single sample. (Default is 200)
        overlap : int, optional
            An `int` that defines the overlap between two consecutive
            samples that are returned when iterating over `_lines`.
            An overlap of 0 means consecutive samples, while an overlap
            of (`sample_size` - 1) means incrementing the samples by one.
            A negative overlap means skipping values between samples. An
            error is raised if `overlap` is larger than or equal to
            `sample_size`. (Default is 0)
        filtered : list, optional
            A list of 'int's of module pair numbers that will be filtered
            from the PEPT data. For use when noisy data affects tracking.
            (Default is [])
        verbose : bool, optional
            An option that enables printing the time taken for the
            initialisation of an instance of the class. Useful when
            reading large files (10gb files for PEPT data is not unheard
            of). (Default is True)
        '''

        if verbose:
            start = time.time()

        if sample_size != 0 and overlap >= sample_size:
            raise ValueError(textwrap.fill((
                f'\n[ERROR]: overlap = {overlap} must be smaller than '
                f'sample_size = {sample_size}\n'
            )))

        self._index = 0
        self._sample_size = sample_size
        self._overlap = overlap

        self.data_file = data_file
        self.filtered = []

        x = 10

        header_buffer_size = 1000

        self.n_events = 0

        if sample_size == 0:
            sample_size = int(
                (os.path.getsize(self.data_file) - header_buffer_size) / 4
            )   # Set the sample size to the expected number of events
            self.overlap = -1


        # Modular camera data reader requires 'itag' for timing. We will drop
        # this column at the end of initialisation
        # Row: [itag, itime, X1, Y1, Z1, X2, Y2, Z2]

        self._lines = np.zeros([sample_size, 8])

        with open(self.data_file, "rb") as f:
            # Skip over the header and handshake word
            f.seek(header_buffer_size)

            word = f.read(4)

            if word.hex() == 'cefacefa':
                # Skip two words
                word = f.read(4)
                word = f.read(4)

            itime = 0
            itag = 0

            # BufTime = 0
            # nBuf = 0

            while word != b'' and (self.n_events < sample_size):

                word = f.read(4)

                # Handshake word
                if word.hex() == 'cefacefa':
                    # Skip two words
                    word = f.read(4)
                    word = f.read(4)

                if word != b'':
                    word = int.from_bytes(word, "little")

                    self._lines[self.n_events, :] = get_pept_LOR(
                        word, itag, itime
                    )   # C function

                    itag = self._lines[self.n_events, 1]
                    itime = self._lines[self.n_events, 1]

                    self.n_events = self.n_events + 1

                    if (self.n_events % x) == 0:
                        print("Got ", self.n_events, "\n")
                        x = x * 10

        # Remove 'zero' lines
        self._lines = self._lines[np.all(self._lines, axis = 1)]

        # Drop itag column
        self._lines = np.delete(self._lines, 0, axis=1)

        self._number_of_lines = len(self._lines)

        # Call the constructor of the superclass `LineData` to initialise all
        # the inner parameters of the class (_index, etc.)
        LineData.__init__(
            self,
            self._lines,
            sample_size = self._sample_size,
            overlap = self._overlap,
            verbose = False
        )


        if verbose:
            end = time.time()
            print(f"Initialising the PEPT data took {end - start} seconds.\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : fpi.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 16.04.2021


import  time
import  textwrap
from    concurrent.futures  import  ThreadPoolExecutor

import  numpy               as      np

from    tqdm                import  tqdm

import  pept
from    .fpi_ext            import  fpi_ext


class FPI:
    '''FPI is a modern voxel-based tracer-location algorithm that can reliably
    work with unknown numbers of tracers in fast and noisy environments.

    It was successfully used to track fast-moving radioactive tracers in pipe
    flows at the Virginia Commonwealth University. If you use this algorithm in
    your work, please cite the following paper:

        Wiggins C, Santos R, Ruggles A. A feature point identification method
        for positron emission particle tracking with multiple tracers. Nuclear
        Instruments and Methods in Physics Research Section A: Accelerators,
        Spectrometers, Detectors and Associated Equipment. 2017 Jan 21;
        843:22-8.

    Permission was granted explicitly by Dr. Cody Wiggins in March 2021 to
    publish his code in the `pept` library under the GNU v3.0 license.

    Two main methods are provided: `fit_sample` for tracking a single voxel
    space (i.e. a single `pept.Voxels`) and `fit` which tracks all the samples
    encapsulated in a `pept.VoxelData` class *in parallel*.

    Attributes
    ----------
    w: double
        Search range to be used in local maxima calculation. Typical values for
        w are 2 - 5 (lower number for more particles or smaller particle
        separation).

    r: double
        Fraction of peak value used as threshold. Typical values for r are
        usually between 0.3 and 0.6 (lower for more particles, higher for
        greater background noise)

    lld_counts: double, default 0
        A secondary lld to prevent assigning local maxima to voxels with very
        low values. The parameter lld_counts is not used much in practice -
        for most cases, it can be set to zero.

    Methods
    -------
    fit_sample(voxels, as_array = False, verbose = False)
        Use the FPI algorithm to locate a tracer from a single voxellised
        space (i.e. from one sample of LoRs).

    fit(voxel_data, max_workers = None, verbose = True)
        Fit multiple voxellised samples of LoRs (in an instance of
        `pept.VoxelData`) and return the tracked tracer locations.

    Examples
    --------
    A typical workflow would involve reading LoRs from a file, creating a lazy
    `VoxelData` voxellised representation, instantiating an `FPI` class,
    tracking the tracer locations from the LoRs, and plotting them.

    >>> import pept
    >>> from pept.tracking.fpi import FPI

    >>> lors = pept.LineData(...)   # set sample_size and overlap appropriately
    >>> voxels = pept.VoxelData(lors)

    >>> fpi = FPI(w = 3, r = 0.4)
    >>> positions = fpi.fit(voxels) # this is a `pept.PointData` instance

    >>> grapher = PlotlyGrapher()
    >>> grapher.add_points(positions)
    >>> grapher.show()

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.

    '''

    def __init__(
        self,
        w = 3.,
        r = 0.4,
        lld_counts = 0.,
        verbose = False,
    ):
        '''`FPI` class constructor.

        Parameters
        ----------
        w: double
            Search range to be used in local maxima calculation. Typical values
            for w are 2 - 5 (lower number for more particles or smaller
            particle separation).

        r: double
            Fraction of peak value used as threshold. Typical values for r are
            usually between 0.3 and 0.6 (lower for more particles, higher for
            greater background noise)

        lld_counts: double, default 0
            A secondary lld to prevent assigning local maxima to voxels with
            very low values. The parameter `lld_counts` is not used much in
            practice - for most cases, it can be set to zero.

        verbose: bool, default False
            Show extra information on class instantiation.

        '''
        self.w = float(w)
        self.r = float(r)
        self.lld_counts = float(lld_counts)


    def fit_sample(
        self,
        voxels,
        timestamp = 0.,
        as_array = False,
        verbose = False,
    ):
        '''Use the FPI algorithm to locate a tracer from a single voxellised
        space (i.e. from one sample of LoRs).

        A sample of LoRs can be voxellised using the `pept.Voxels.from_lines`
        method before calling this function.

        Parameters
        ----------
        voxels: pept.Voxels
            A single voxellised space (i.e. from a single sample of LoRs) for
            which the tracers' locations will be found using the FPI method.

        timestamp: float, default 0.
            The timestamp to associate with the tracer positions found in this
            voxel space.

        as_array: bool, default False
            If `True`, return the found tracers' locations as a NumPy array.
            Otherwise, return them in a `pept.PointData` instance.

        verbose: bool, default False
            Show extra information on the sample processing step.

        Returns
        -------
        locations: numpy.ndarray or pept.PointData
            The tracked locations found; if `as_array` is True, they are
            returned as a NumPy array with columns [time, x, y, z, error_x,
            error_y, error_z]. If `as_array` is False, the points are returned
            in a `pept.PointData` for ease of visualisation.

        Raises
        ------
        TypeError
            If `voxels` is not an instance of `pept.Voxels` (or subclass
            thereof).
        '''
        if not isinstance(voxels, pept.Voxels):
            raise TypeError(textwrap.fill((
                "The input `voxels` must be an instance of `pept.Voxels`. "
                f"Received `{type(voxels)}`."
            )))

        points = fpi_ext(
            voxels.voxels,
            self.w,
            self.r,
            self.lld_counts,
        )

        # Insert the time column and translate the coordinates from the voxel
        # space to the physical space
        points = np.insert(points, 0, timestamp, axis = 1)
        points[:, 1:4] *= voxels.voxel_size
        points[:, 1:4] += [voxels.xlim[0], voxels.ylim[0], voxels.zlim[0]]

        if as_array:
            return points

        return pept.PointData(points)


    def _fit_voxel_data(self, voxel_data, index):
        # The voxellisation step is only computed when indexing `voxel_data`
        voxels = voxel_data[index]
        points = fpi_ext(voxels, self.w, self.r, self.lld_counts)

        # Insert the time column and translate the coordinates from the voxel
        # space to the physical space
        timestamp = voxel_data.line_data[index][:, 0].mean()
        points = np.insert(points, 0, timestamp, axis = 1)

        points[:, 1:4] *= voxels.voxel_size
        points[:, 1:4] += [voxels.xlim[0], voxels.ylim[0], voxels.zlim[0]]

        return points


    def fit(
        self,
        voxel_data,
        max_workers = None,
        verbose = True,
    ):
        '''Fit multiple voxellised samples of LoRs (in an instance of
        `pept.VoxelData`) and return the tracked tracer locations.

        The input `pept.VoxelData` class can voxellise samples of LoRs (from a
        `pept.LineData`) in parallel, *on demand* - that is, the voxels will
        not be stored in memory all at once, only when needed.

        This is a convenience function that asynchronously iterates through the
        samples in a `VoxelData`, finding the tracer locations. For more
        fine-grained control over the tracking, the `fit_sample` method can be
        used for individual samples.

        Parameters
        ----------
        voxel_data: pept.VoxelData
            The voxellised samples of LoRs; the `pept.VoxelData` class can
            compute the voxellisation steps in parallel, *on-demand*,
            minimising the amount of memory needed at once.

        max_workers: int, optional
            The maximum number of threads that will be used for asynchronously
            finding the tracers' positions. If unset (`None`), the maximum
            number of threads available on the machine will be used.

        verbose : bool, default True
            Provide extra information when tracking: time the operation and
            show a progress bar.

        Returns
        -------
        locations : pept.PointData
            The tracer locations found.

        Raises
        ------
        TypeError
            If the input `voxel_data` is not an instance of `pept.VoxelData`
            (or subclass thereof).

        '''
        if not isinstance(voxel_data, pept.VoxelData):
            raise TypeError(textwrap.fill((
                "The input `voxel_data` must be an instance of "
                f"`pept.VoxelData`. Received `{type(voxel_data)}`."
            )))

        with ThreadPoolExecutor(max_workers = max_workers) as executor:
            futures = []

            for index in range(len(voxel_data)):
                futures.append(executor.submit(
                    self._fit_voxel_data,
                    voxel_data,
                    index,
                ))

            if verbose:
                futures = tqdm(futures)
                time_start = time.time()

            points = [f.result() for f in futures]

            if verbose:
                time_end = time.time()
                print(f"Fitted data in {time_end - time_start} s")

        points = np.vstack(points)

        return pept.PointData(points)


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"w =           {self.w}\n"
            f"r =           {self.r}\n"
            f"lld_counts =  {self.lld_counts}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL
        docstr = (
            "Class instance that inherits from `FPI`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n"
            "----------\n"
            f"{self.__str__()}\n"
        )

        return docstr



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : fpi_ext.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 06.04.2020


# distutils: language=c++

# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np      # import numpy for Python functions
cimport numpy as np     # import numpy for C functions (numpy's C API)


np.import_array()


cdef extern from "calcPosFPI.hpp":
    double* calcPosFPIC(
        double *voxels,
        Py_ssize_t length,
        Py_ssize_t width,
        Py_ssize_t depth,
        double w,
        double r,
        double lldCounts,
        Py_ssize_t *out_rows,
        Py_ssize_t *out_cols
    ) nogil


cpdef np.ndarray[double, ndim=2] fpi_ext(
    double[:, :, :] voxels,
    double w,
    double r,
    double lldCounts = 0.,
):
    '''Find a tracer's location from a pre-computed voxellised sample of LoRs
    using the Feature Point Identification (FPI) method.

    ::

        Function signature:

            np.ndarray[double, ndim=2] fpi_ext(
                double[:, :, :] voxels,
                double w,
                double r,
                double lldCounts,
            )

    FPI is a modern tracer-location algorithm that was successfully used to
    track fast-moving radioactive tracers in pipe flows at the Virginia
    Commonwealth University. If you use this algorithm in your work, please
    cite the following paper:

        Wiggins C, Santos R, Ruggles A. A feature point identification method
        for positron emission particle tracking with multiple tracers. Nuclear
        Instruments and Methods in Physics Research Section A: Accelerators,
        Spectrometers, Detectors and Associated Equipment. 2017 Jan 21;
        843:22-8.

    Permission was granted explicitly by Dr. Cody Wiggins in March 2021 to
    publish his code in the `pept` library under the GNU v3.0 license.

    The points returned by this function are in *voxel dimensions*, without
    timestamps. They can be translated into physical dimensions and timestamps
    can be added after calling it, e.g. with the `pept.tracking.fpi.FPI`
    class.

    Parameters
    ----------
    voxels: (L, W, D) numpy.ndarray[ndim = 2, dtype = numpy.float64]
        The 3D grid of voxels, initialised to zero. It can be created with
        `numpy.zeros((length, width, depth))`.

    w: double
        Search range to be used in local maxima calculation. Typical values for
        w are 2 - 5 (lower number for more particles or smaller particle
        separation).

    r: double
        Fraction of peak value used as threshold. Typical values for r are
        usually between 0.3 and 0.6 (lower for more particles, higher for
        greater background noise)

    lldCounts: double, default 0
        A secondary lld to prevent assigning local maxima to voxels with very
        low values. The parameter lldCounts is not used much in practice -
        for most cases, it can be set to zero.

    '''

    cdef double *points
    cdef Py_ssize_t nrows = 0
    cdef Py_ssize_t ncols = 0
    cdef np.npy_intp[2] size

    with nogil:
        points = calcPosFPIC(
            &voxels[0, 0, 0],
            voxels.shape[0],
            voxels.shape[1],
            voxels.shape[2],
            w,
            r,
            lldCounts,
            &nrows,
            &ncols,
        )

    size[0] = nrows
    size[1] = ncols

    # Use the `minpoints` pointer as the internal data of a numpy array
    cdef extern from "numpy/arrayobject.h":
        void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

    cdef np.ndarray[double, ndim=2] points_arr = np.PyArray_SimpleNewFromData(
        2, size, np.NPY_FLOAT64, points
    )
    PyArray_ENABLEFLAGS(points_arr, np.NPY_OWNDATA)

    return points_arr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.08.2019


'''The `peptml` package implements an efficient, optionally-parallel
hierarchical density-based clustering algorithm for general Positron Emission
Particle Tracking (PEPT).

Summary
-------
A typical workflow for using the `peptml` package would be:

1. Read the LoRs into a `pept.LineData` class instance and set the
   `sample_size` and `overlap` appropriately.
2. Compute the cutpoints using the `pept.tracking.peptml.Cutpoints` class.
3. Instantiate an `pept.tracking.peptml.HDBSCANClusterer` class and cluster the
   cutpoints found previously.
4. Optional: cluster the results from the previous step again for smoother,
   tighter trajectories.

Extended Summary
----------------
The PEPT-ML algorithm [1]_ works using the following steps:

1. Split the data into a series of individual "samples", each containing a
   given number of LoRs. Use the base class pept.LineData for this.
2. For every sample of LoRs, compute the *cutpoints*, or the points in space
   that minimise the distance to every pair of lines.
3. Cluster every sample using HDBSCAN and extract the centres of the clusters
   ("1-pass clustering").
4. Split the centres into samples of a given size.
5. Cluster every sample of centres using HDBSCAN and extract the centres of the
   clusters ("2-pass clustering").
6. Construct the trajectory of every particle using the centres from the
   previous step.

More tutorials and examples can be found on the University of Birmingham
Positron Imaging Centre's GitHub repository.

PEPT-ML was successfuly used at the University of Birmingham to analyse real
Fluorine-18 tracers in air.

Modules Provided
----------------

::

    pept.tracking.peptml
    │
    Functions imported into the subpackage root:
    ├── find_cutpoints :        Find cutpoints from a NumPy array of lines.
    ├── find_minpoints :        Find MDPs of combinations from a line array.
    ├── get_cutoffs :           Find cutpoint cutoffs from an array of lines.
    │
    Classes imported into the subpackage root:
    ├── Cutpoints :         Compute cutpoints from samples in a `LineData`.
    ├── Minpoints :         Compute minpoints from samples in a `LineData`.
    └── HDBSCANClusterer :  Cluster samples of cutpoints in parallel.

References
----------
.. [1] Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
   using machine learning. Review of Scientific Instruments.
   2020 Jan 1;91(1):013329.
   https://doi.org/10.1063/1.5129251
'''


from    .cutpoints  import  find_cutpoints
from    .minpoints  import  find_minpoints
from    .cutpoints  import  get_cutoffs
from    .cutpoints  import  Cutpoints
from    .minpoints  import  Minpoints

from    .peptml     import  HDBSCANClusterer


__all__ = [
    "find_cutpoints",
    "find_minpoints",
    "get_cutoffs",
    "Cutpoints",
    "Minpoints",
    "HDBSCANClusterer"
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : test_base_classes.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.08.2019

#from    nose    import  with_setup

from    nose.tools  import  raises

import  pept
import  numpy   as      np


def setup_module(module):
    print ("") # this is to get a newline after the dots
    print ("setup_module before anything in this file")

def teardown_module(module):
    print ("teardown_module after everything in this file")


class TestLineData:

    # Run when class is created (before running any method)
    @classmethod
    def setup_class(cls):
        cls.good_data = np.arange(2800).reshape(400, 7)
        cls.bad_data = np.arange(3500).reshape(5, 100, 7)


    # Run when class is destroyed (after running all methods)
    @classmethod
    def teardown_class(cls):
        pass


    # Run before each class method
    def setup(self):
        pass


    # Run after each class method
    def teardown(self):
        pass


    def test_good_data(self):
        samples = pept.LineData(self.good_data, sample_size = 200, overlap = 10, verbose = False)
        # Test private attributes
        assert samples._index == 0, "_index was not set to 0"
        assert samples._overlap == 10, "_overlap was not set correctly"
        assert samples._sample_size == 200, "_sample_size was not set correctly"
        assert np.array_equal(samples._lines, self.good_data) == True
        assert samples._lines.flags['C_CONTIGUOUS'] == True, "_lines is not C-contiguous"
        assert samples._number_of_lines == len(samples._lines), "_number_of_lines was not set correctly"

        # Test properties
        assert np.array_equal(samples.lines, samples._lines) == True
        assert samples.sample_size == samples._sample_size
        assert samples.overlap == samples._overlap
        assert samples.number_of_samples == 2, "number of samples was not calculated correctly"
        assert samples.number_of_lines == samples._number_of_lines

        # Test property setters
        samples.sample_size = 300
        assert samples.sample_size == 300
        assert samples._index == 0
        samples.sample_size = 200

        samples.overlap = 50
        assert samples.overlap == 50
        assert samples._index == 0
        samples.overlap = 10


    @raises(ValueError)
    def test_error_sample_size(self):
        samples = pept.LineData(self.good_data, sample_size = 200, overlap = 100, verbose = False)
        # Should not be able to set sample_size <= overlap
        samples.sample_size = 100


    @raises(ValueError)
    def test_error_overlap(self):
        samples = pept.LineData(self.good_data, sample_size = 200, overlap = 100, verbose = False)
        # Should not be able to set sample_size <= overlap
        samples.overlap = 200


    def test_sample_n(self):
        pass





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : setup.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev


import  io
import  os
import  sys
import  warnings
from    shutil              import  rmtree

from    setuptools          import  find_packages, setup, Command, Extension


try:
    import  numpy               as      np
    from    Cython.Build        import  cythonize
    from    Cython.Distutils    import  build_ext
except ImportError as e:
    warnings.warn(e.args[0])
    warnings.warn((
        'The pept package requires Cython and Numpy to be pre-installed'
    ))
    raise ImportError((
        'Cython or Numpy not found! Please install Cython and Numpy (or run '
        '`pip install -r requirements.txt`) and try again.'
    ))


# Package meta-data.
NAME = 'pept'
DESCRIPTION = (
    'A Python library that unifies Positron Emission Particle Tracking (PEPT) '
    'research, including tracking, simulation, data analysis and '
    'visualisation tools.'
)
URL = 'https://github.com/uob-positron-imaging-centre/pept'
EMAIL = 'a.l.nicusan@bham.ac.uk'
AUTHOR = 'Andrei Leonard Nicusan'
REQUIRES_PYTHON = '>=3.6.0'


def requirements(filename):
    # The dependencies are the same as the contents of requirements.txt
    with open(filename) as f:
        return [line.strip() for line in f if line.strip()]


# What packages are required for this module to be executed?
REQUIRED = requirements('requirements.txt')


# What packages are optional?
EXTRAS = {
    'docs': requirements('requirements_docs.txt'),
}

# Cythonize keyword arguments
cythonize_kw = dict(language_level = 3)

# Compiler arguments for each extension - with *full* optimisations.
# Unix-specific compiler args are followed by MSVC ones; they will be filtered
# based on the compiler used in `BuildExtCompilerSpecific`
cy_extension_kw = dict()
cy_extension_kw['extra_compile_args'] = [
    '-Ofast', '-flto', '/O2', '/fp:fast', '/GL'
]
cy_extension_kw['extra_link_args'] = ['-flto', '/LTCG']
cy_extension_kw['include_dirs'] = [np.get_include()]

# Compiler arguments for each extension - with *strict floating point*
# optimisations.
cy_extension_kw_strict = dict()
cy_extension_kw_strict['extra_compile_args'] = [
    '-O3', '-flto', '/O2', '/GL'
]
cy_extension_kw_strict['extra_link_args'] = ['-flto', '/LTCG']
cy_extension_kw_strict['include_dirs'] = [np.get_include()]

cy_extensions = [
    Extension(
        'pept.scanners.parallel_screens.extensions.binary_converter',
        ['pept/scanners/parallel_screens/extensions/binary_converter.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.utilities.cutpoints.find_cutpoints',
        ['pept/utilities/cutpoints/find_cutpoints.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.utilities.cutpoints.find_minpoints',
        ['pept/utilities/cutpoints/find_minpoints.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.processing.circles_ext',
        ['pept/processing/circles_ext.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.processing.occupancy_ext',
        ['pept/processing/occupancy_ext.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.scanners.modular_camera.extensions.get_pept_event',
        ['pept/scanners/modular_camera/extensions/get_pept_event.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.utilities.traverse.traverse3d',
        ['pept/utilities/traverse/traverse3d.pyx'],
        **cy_extension_kw_strict
    ),
    Extension(
        'pept.utilities.traverse.traverse2d',
        ['pept/utilities/traverse/traverse2d.pyx'],
        **cy_extension_kw_strict
    ),
    Extension(
        'pept.tracking.trajectory_separation.distance_matrix_reachable',
        ['pept/tracking/trajectory_separation/distance_matrix_reachable.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.tracking.birmingham_method.extensions.birmingham_method',
        ['pept/tracking/birmingham_method/extensions/birmingham_method.pyx'],
        **cy_extension_kw
    ),
    Extension(
        'pept.tracking.fpi.fpi_ext',
        ['pept/tracking/fpi/fpi_ext.pyx'],
        **cy_extension_kw
    ),
]

extensions = cythonize(cy_extensions, **cythonize_kw)


class BuildExtCompilerSpecific(build_ext):
    '''Before compiling extensions, ensure only valid compiler arguments are
    used - e.g. MSVC expects "/O2", while GCC and Clang expect "-O3".
    '''
    def build_extensions(self):
        # If compiling under MSVC, remove the Unix-specific compiler arguments
        if "msvc" in self.compiler.compiler_type.lower():
            for ext in self.extensions:
                ext.extra_compile_args = [
                    ca for ca in ext.extra_compile_args if ca.startswith("/")
                ]

                ext.extra_link_args = [
                    la for la in ext.extra_link_args if la.startswith("/")
                ]

        # Otherwise remove the MSVC-specific compiler arguments
        else:
            for ext in self.extensions:
                ext.extra_compile_args = [
                    ca for ca in ext.extra_compile_args if ca.startswith("-")
                ]

                ext.extra_link_args = [
                    la for la in ext.extra_link_args if la.startswith("-")
                ]

        build_ext.build_extensions(self)


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Load the package's __version__.py module as a dictionary.
about = {}
project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
with open(os.path.join(here, project_slug, '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name = NAME,
    version = about['__version__'],
    description = DESCRIPTION,
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = AUTHOR,
    author_email = EMAIL,
    python_requires = REQUIRES_PYTHON,
    url = URL,
    packages = find_packages(
        exclude = ["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    install_requires = REQUIRED,
    extras_require = EXTRAS,
    include_package_data = True,
    keywords = 'pept positron emission particle tracking',
    license = 'GNU',
    classifiers = [
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Cython',
        'Programming Language :: C',
        'Programming Language :: C++',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # $ setup.py publish support.
    cmdclass = {
        'upload': UploadCommand,
        'build_ext': BuildExtCompilerSpecific,
    },
    ext_modules = extensions,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : trajectory_separation.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019


import  os

import  numpy                       as      np
from    scipy.spatial               import  cKDTree
from    scipy.sparse.csgraph        import  minimum_spanning_tree

import  pept
import  hdbscan

from    .tco                        import  with_continuations
from    .distance_matrix_reachable  import  distance_matrix_reachable




def trajectory_errors(
    true_positions,
    tracked_positions,
    averages = True,
    stds = False,
    errors = False,
    max_workers = None
):

    # true_positions, tracked_positions cols: [t, x, y, z, ...]
    true_positions = np.asarray(true_positions, dtype = float)
    tracked_positions = np.asarray(tracked_positions, dtype = float)

    # Construct k-d tree for fast nearest neighbour lookup.
    tree = cKDTree(true_positions[:, 1:4])

    # errors cols: [err, err_x, err_y, err_z]
    errors = np.zeros((len(tracked_positions), 4))

    if max_workers is None:
        max_workers = os.cpu_count()

    # Find closest true point to each tracked point.
    for i, pos in enumerate(tracked_positions[:, 1:4]):
        dist, index = tree.query(pos, k = 1, n_jobs = max_workers)

        errors[i, 0] = np.linalg.norm(pos - true_positions[index])
        errors[i, 1:4] = np.abs(pos - true_positions[index, 0:3])
        # errors[i, 1] = np.abs(pos[0] - true_positions[index, 0])
        # errors[i, 2] = np.abs(pos[1] - true_positions[index, 1])
        # errors[i, 3] = np.abs(pos[2] - true_positions[index, 2])

    # Append and return the results that were asked for.
    results = []

    if averages:
        results.append(np.average(errors, axis = 0))
    if stds:
        results.append(np.std(errors, axis = 0))
    if errors:
        results.append(errors)

    # If a single result was asked for, return it directly; otherwise, use a
    # tuple.
    if len(results) == 1:
        return results[0]
    else:
        return tuple(results)




def segregate_trajectories(
    point_data,
    points_window,
    trajectory_cut_distance,
    min_trajectory_size = 5,
    as_list = False,
    return_mst = False
):
    '''Segregate the intertwined points from multiple trajectories into
    individual paths.

    The points in `point_data` (a numpy array or `pept.PointData`) are used to
    construct a minimum spanning tree in which every point can only be
    connected to `points_window` points around it - this "window" refers to the
    points in the initial data array, sorted based on the time column;
    therefore, only points within a certain timeframe can be connected. All
    edges (or "connections") in the minimum spanning tree that are larger than
    `trajectory_cut_distance` are removed (or "cut") and the remaining
    connected "clusters" are deemed individual trajectories if they contain
    more than `min_trajectory_size` points.

    The trajectory indices (or labels) are appended to `point_data`. That is,
    for each data point (i.e. row) in `point_data`, a label will be appended
    starting from 0 for the corresponding trajectory; a label of -1 represents
    noise. If `point_data` is a numpy array, a new numpy array is returned; if
    it is a `pept.PointData` instance, a new instance is returned.

    This function uses single linkage clustering with a custom metric for
    spatio-temporal data to segregate trajectory points. The single linkage
    clustering was optimised for this use-case: points are only connected if
    they are within a certain `points_window` in the time-sorted input array.
    Sparse matrices are also used for minimising the memory footprint.

    Parameters
    ----------
    point_data : (M, N>=4) numpy.ndarray or pept.PointData
        The points from multiple trajectories. Each row in `point_data` will
        have a timestamp and the 3 spatial coordinates, such that the data
        columns are [time, x_coord, y_coord, z_coord]. Note that `point_data`
        can have more data columns and they will simply be ignored.
    points_window : int
        Two points are "reachable" (i.e. they can be connected) if and only if
        they are within `points_window` in the time-sorted input `point_data`.
        As the points from different trajectories are intertwined (e.g. for two
        tracers A and B, the `point_data` array might have two entries for A,
        followed by three entries for B, then one entry for A, etc.), this
        should optimally be the largest number of points in the input array
        between two consecutive points on the same trajectory. If
        `points_window` is too small, all points in the dataset will be
        unreachable. Naturally, a larger `time_window` correponds to more pairs
        needing to be checked (and the function will take a longer to
        complete).
    trajectory_cut_distance : float
        Once all the closest points are connected (i.e. the minimum spanning
        tree is constructed), separate all trajectories that are further
        apart than `trajectory_cut_distance`.
    min_trajectory_size : float, default 5
        After the trajectories have been cut, declare all trajectories with
        fewer points than `min_trajectory_size` as noise.
    as_list : bool, default False
        If True, return a list of arrays, where each array contains the points
        in a single trajectory. In other words, return separate, single
        trajectories in a list. If False, return a single array of all points
        (if `point_data` was a `numpy.ndarray`) or a `pept.PointData`
        (if `point_data` was a `pept.PointData` instance).
    return_mst : bool, default False
        If `True`, the function will also return the minimum spanning tree
        constructed using the input `point_data`. This is a numpy array with
        columns [vertex1, vertex2, edge_length], where vertex1 and vertex2 are
        the indices in `point_data` of the connected points, and edge_length is
        the euclidian distance between them.

    Returns
    -------
    points_labelled: numpy.ndarray or pept.PointData or list of numpy.ndarray
        If `as_list` is `False`, this is the `point_data` array or
        `pept.PointData` instance with an extra column for the trajectory index
        (i.e. label) - the return type is similar to the input type. If
        `as_list` is `True`, this is a list of arrays, in which each array
        contains the points in a single_trajectory; these still include the
        trajectory label. A label value of `-1` indicates noise; the found
        trajectories are then labelled starting from 0.
    mst: numpy.ndarray, optional
        If `return_mst` is `True`, another numpy array is returned as a second
        variable containing the columns [vertex1, vertex2, edge_length], where
        vertex1 and vertex2 are the indices in `point_data` of the connected
        points, and edge_length is the euclidian distance between them.

    Raises
    ------
    ValueError
        If `point_data` is a numpy array with fewer than 4 columns.
    ValueError
        If `points_window` is smaller than 1.

    Examples
    --------
    A typical workflow would involve transforming LoRs into points using some
    tracking algorithm. These points include all tracers moving through the
    system, being intertwined (e.g. for two tracers A and B, the `point_data`
    array might have two entries for A, followed by three entries for B, then
    one entry for A, etc.). They can be segregated based on position alone
    using this function; take for example two tracers that go downwards (below,
    'x' is the position, and in parens is the array index at which that point
    is found).

    ::

        `points`, numpy.ndarray, shape (10, 4), columns [time, x, y, z]:
            x (1)                       x (2)
             x (3)                     x (4)
               x (5)                 x (7)
               x (6)                x (9)
              x (8)                 x (10)

    >>> import pept.tracking.trajectory_separation as tsp
    >>> points_window = 10
    >>> trajectory_cut_distance = 15    # mm
    >>> segregated_trajectories = tsp.segregate_trajectories(
    >>>     points, points_window, trajectory_cut_distance
    >>> )

    ::

        `segregated_trajectories`, numpy.ndarray, shape (10, 5),
        columns [time, x, y, z, trajectory_label]:
            x (1, label = 0)            x (2, label = 1)
             x (3, label = 0)          x (4, label = 1)
               x (5, label = 0)      x (7, label = 1)
               x (6, label = 0)     x (9, label = 1)
              x (8, label = 0)      x (10, label = 1)

    See Also
    --------
    connect_trajectories : Connect segregated trajectories based on tracer
                           signatures.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    # Check `point_data` is a numpy array or pept.PointData
    if isinstance(point_data, pept.PointData):
        pts = point_data.points
    else:
        pts = np.asarray(point_data)
        if pts.ndim != 2 or pts.shape[1] < 4:
            raise ValueError((
                "\n[ERROR]: `point_data` should have dimensions (M, N), where "
                f"N >= 4. Received {pts.shape}.\n"
            ))

    # Sort pts based on the time column (col 0) and create a C-ordered copy to
    # send to Cython.
    pts = np.asarray(pts[pts[:, 0].argsort()], dtype = float, order = "C")

    # Type-check the input parameters
    points_window = int(points_window)
    if points_window < 1:
        raise ValueError((
            "\n[ERROR]: `points_window` should be larger than 1! Received "
            f"{points_window}.\n"
        ))

    trajectory_cut_distance = float(trajectory_cut_distance)
    min_trajectory_size = int(min_trajectory_size)
    return_mst = bool(return_mst)

    # Calculate the sparse distance matrix between reachable points. This is an
    # optimised Cython function returning a sparse CSR matrix.
    distance_matrix = distance_matrix_reachable(pts, points_window)

    # Construct the minimum spanning tree from the sparse distance matrix. Note
    # that `mst` is also a sparse CSR matrix.
    mst = minimum_spanning_tree(distance_matrix)

    # Get the minimum spanning tree edges into the [vertex 1, vertex 2,
    # edge distance] format, then sort it based on the edge distance.
    mst = mst.tocoo()
    mst_edges = np.vstack((mst.row, mst.col, mst.data)).T
    mst_edges = mst_edges[mst_edges[:, 2].argsort()]

    # Create the single linkage tree from the minimum spanning tree edges using
    # internal hdbscan methods (because they're damn fast). This should be a
    # fairly quick step.
    single_linkage_tree = hdbscan._hdbscan_linkage.label(mst_edges)
    single_linkage_tree = hdbscan.plots.SingleLinkageTree(single_linkage_tree)

    # Cut the single linkage tree at `trajectory_cut_distance` and get the
    # cluster labels, setting clusters smaller than `min_trajectory_size` to
    # -1 (i.e. noise).
    labels = single_linkage_tree.get_clusters(
        trajectory_cut_distance,
        min_trajectory_size
    )

    # Append the labels to `pts`.
    pts = np.append(pts, labels[:, np.newaxis], axis = 1)

    # Returns based on as_list, return_mst and input data type
    if as_list:
        # Get a list of arrays for each trajectory
        separate_pts = pept.utilities.group_by_column(pts, -1)
        if return_mst:
            return separate_pts, mst_edges
        return separate_pts

    # If `point_data` was a `pept.PointData` instance, return a new
    # `pept.PointData` with the new label column.
    if isinstance(point_data, pept.PointData):
        point_data_labelled = pept.PointData(
            pts,
            sample_size = point_data.sample_size,
            overlap = point_data.overlap,
            verbose = False
        )
        if return_mst:
            return point_data_labelled, mst_edges
        else:
            return point_data_labelled
    elif return_mst:
        return pts, mst_edges
    return pts




def connect_trajectories(
    trajectories_points,
    max_time_difference,
    max_signature_difference,
    points_to_check = 50,
    signature_col = 4,
    label_col = -1,
    as_list = False
):
    '''Connect segregated trajectories based on tracer signatures.

    A pair of trajectories in `trajectories_points` will be connected if their
    ends have a timestamp difference that is smaller than `max_time_difference`
    and the difference between the signature averages of the closest
    `points_to_check` points is smaller than `max_signature_difference`.

    The `trajectories_points` are distinguished based on the trajectory
    indices in the data column `label_col`. This can be achieved using the
    `segregate_trajectories` function, which appends the labels to the data
    points.

    Because the tracer signature (e.g. cluster size in PEPT-ML) varies with the
    tracer position in the system, an average of `points_to_check` points
    is used for connecting pairs of trajectories.

    Parameters
    ----------
    trajectories_points : (M, N>=6) numpy.ndarray or pept.PointData
        A numpy array of points that have a timestamp, spatial coordinates,
        a tracer signature (such as cluster size in PEPT-ML) and a trajectory
        index (or label). The data columns in `trajectories_points` are then
        [time, x, y, z, ..., signature, ..., label, ...]. Note that the
        timestamps and spatial coordinates must be the first 4 columns, while
        the signature and label columns may be anywhere and are pointed at
        by `signature_col` and `label_col`.
    max_time_difference : float
        Only try to connect trajectories whose ends have a timestamp difference
        smaller than `max_time_difference`.
    max_signature_difference : float
        Connect two trajectories if the difference between the signature
        averages of the closest `points_to_check` is smaller than this.
    points_to_check : int, default 50
        The number of points used when computing the average tracer signature
        in one trajectory.
    signature_col : int, default 4
        The column in `trajectories_points` that contains the tracer
        signatures. The default is 4 (i.e. the signature comes right after
        the spatial coordinates).
    label_col : int, default -1
        The column in `trajectories_points` that contains the trajectory
        indices (labels). The default is -1 (i.e. the last column).
    as_list : bool, default False
        If True, return a list of arrays, where each array contains the points
        in a single trajectory. In other words, return separate, single
        trajectories in a list. If False, return a single array of all points
        (if `trajectories_points` was a `numpy.ndarray`) or a `pept.PointData`
        (if `trajectories_points` was a `pept.PointData` instance), but with
        labels changed to reflect the connected trajectories.

    Returns
    -------
    numpy.ndarray or pept.PointData or list of numpy.ndarray
        If `as_list` is True, return separate, single trajectories in a list.
        If `as_list` is False, return a single array of all points
        (if `trajectories_points` was a `numpy.ndarray`) or a `pept.PointData`
        (if `trajectories_points` was a `pept.PointData` instance), but with
        labels changed to reflect the connected trajectories.

    Raises
    ------
    ValueError
        If `point_data` is a numpy array with fewer than 6 columns.

    Notes
    -----
    The labels are changed in-place to reflect the connected trajectories. For
    example, if there are 3 trajectories with labels 0, 1, 2 and the first two
    are connected, then all points which previously had the label 1 will be
    changed to label 0; the last trajectory's label remains unchanged, 2.

    Examples
    --------
    [TODO] - add full tutorial page on Bham PIC GitHub page for this.

    See Also
    --------
    segregate_trajectories : Segregate the intertwined points from multiple
                             trajectories into individual paths.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    # Check `point_data` is a numpy array or pept.PointData
    if isinstance(trajectories_points, pept.PointData):
        trajs = trajectories_points.points
    else:
        trajs = np.asarray(trajectories_points, dtype = float, order = "C")
        if trajs.ndim != 2 or trajs.shape[1] < 6:
            raise ValueError((
                "\n[ERROR]: `trajectories_points` should have dimensions "
                f"(M, N), where N >= 6. Received {trajs.shape}.\n"
            ))

    # Type-check the input parameters
    max_time_difference = float(max_time_difference)
    max_signature_difference = float(max_signature_difference)
    points_to_check = int(points_to_check)
    signature_col = int(signature_col)
    label_col = int(label_col)
    as_list = bool(as_list)

    # Separate the trajs array into a list of individual trajectories based on
    # the `label_col`.
    trajectory_list = pept.utilities.group_by_column(trajs.copy(), label_col)

    trajectory_list = _connect_trajectories(
        trajectory_list,
        max_time_difference,
        max_signature_difference,
        points_to_check,
        signature_col,
        label_col
    )

    if as_list:
        return trajectory_list
    elif isinstance(trajectories_points, pept.PointData):
        trajectories_points_connected = pept.PointData(
            np.vstack(trajectory_list),
            sample_size = trajectories_points.sample_size,
            overlap = trajectories_points.overlap,
            verbose = False
        )
        return trajectories_points_connected
    else:
        return np.vstack(trajectory_list)




@with_continuations()       # Use tail-call optimisation from tco.py
def _connect_trajectories(
    trajectory_list,
    max_time_difference,
    max_signature_difference,
    points_to_check,
    signature_col = 4,
    label_col = -1,
    self = None
):
    number_of_trajectories = len(trajectory_list)

    # Check all pairs of trajectories. Each trajectory has two ends:
    # Traj1: [End11, End12]
    # Traj2: [End21, End22]
    for i in range(number_of_trajectories - 1):
        for j in range(i + 1, number_of_trajectories):

            t1 = trajectory_list[i]     # Traj1
            t2 = trajectory_list[j]     # Traj2

            # Try to connect End11 with End22
            # Check the time difference between End11 and End22 is small enough
            if time_difference(t1, t2) < max_time_difference:

                # Check the signature difference of the last `points_to_check`
                # points is small enough
                if signature_difference(
                    t1, t2, points_to_check, signature_col
                ) < max_signature_difference:

                    # Assimilate the column labels
                    t1[:, label_col] = t2[0, label_col]

                    # Merge the two trajectories
                    connected_trajectory = np.append(t2, t1, axis = 0)

                    # Delete the individual trajectories and append the merged
                    # trajectories to `trajectory_list`
                    del trajectory_list[i], trajectory_list[j - 1]
                    trajectory_list.append(connected_trajectory)

                    # Call the function again (to reinitialise
                    # `number_of_trajectories`) using tail call optimisation
                    # from tco.py
                    return self(
                        trajectory_list,
                        max_time_difference,
                        max_signature_difference,
                        points_to_check,
                        signature_col,
                        label_col
                    )

            # Try to connect End12 with End21
            elif time_difference(t2, t1) < max_time_difference:

                if signature_difference(
                    t2, t1, points_to_check, signature_col
                ) < max_signature_difference:

                    t2[:, label_col] = t1[0, label_col]
                    connected_trajectory = np.append(t1, t2, axis = 0)

                    del trajectory_list[i], trajectory_list[j - 1]
                    trajectory_list.append(connected_trajectory)

                    return self(
                        trajectory_list,
                        max_time_difference,
                        max_signature_difference,
                        points_to_check,
                        signature_col,
                        label_col
                    )

    return trajectory_list




def signature_difference(
    traj1,
    traj2,
    points_to_check,
    signature_col
):
    return np.abs(
        np.average(traj1[:points_to_check, signature_col]) -
        np.average(traj2[-points_to_check:, signature_col])
    )




def time_difference(traj1, traj2):
    return np.abs(traj1[0, 0] - traj2[-1, 0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 01.04.2021


from    .binary_converter   import  convert_adac_forte


__all__ = [
    "convert_adac_forte",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : birmingham_method.py
# License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


import  os
import  time
import  textwrap
import  warnings
from    concurrent.futures              import  ThreadPoolExecutor

import  numpy                           as      np
from    tqdm                            import  tqdm

import  pept

from    .extensions.birmingham_method   import  birmingham_method


class BirminghamMethod:
    '''The Birmingham Method is an efficient, analytical technique for tracking
    tracers using the LoRs from PEPT data.

    Two main methods are provided: `fit_sample` for tracking a single numpy
    array of LoRs (i.e. a single sample) and `fit` which tracks all the samples
    encapsulated in a `pept.LineData` class *in parallel*.

    For the given `sample` of LoRs (a numpy.ndarray), this function minimises
    the distance between all of the LoRs, rejecting a fraction of lines that
    lie furthest away from the calculated distance. The process is repeated
    iteratively until a specified fraction (`fopt`) of the original subset of
    LORs remains.

    This class is a wrapper around the `birmingham_method` subroutine
    (implemented in C), providing tools for asynchronously tracking samples of
    LoRs. It can return `PointData` classes which can be easily manipulated and
    visualised.

    Attributes
    ----------
    fopt : float
        Floating-point number between 0 and 1, representing the target fraction
        of LoRs in a sample used to locate a tracer.

    Methods
    -------
    fit_sample(sample, get_used = False, as_array = True, verbose = False)
        Use the Birmingham method to track a tracer location from a numpy
        array (i.e. one sample) of LoRs.

    fit(line_data, max_error = 10, get_used = False, max_workers = None,\
        verbose = True)
        Fit lines of response (an instance of 'LineData') and return the
        tracked locations and (optionally) the LoRs that were used.

    Examples
    --------
    A typical workflow would involve reading LoRs from a file, instantiating a
    `BirminghamMethod` class, tracking the tracer locations from the LoRs, and
    plotting them.

    >>> import pept
    >>> from pept.tracking.birmingham_method import BirminghamMethod

    >>> lors = pept.LineData(...)   # set sample_size and overlap appropriately
    >>> bham = BirminghamMethod()
    >>> locations = bham.fit(lors)  # this is a `pept.PointData` instance

    >>> grapher = PlotlyGrapher()
    >>> grapher.add_points(locations)
    >>> grapher.show()

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    pept.scanners.ParallelScreens : Initialise a `pept.LineData` instance from
                                    parallel screens PEPT detectors.
    '''

    def __init__(self, fopt = 0.5, verbose = False):
        '''`BirminghamMethod` class constructor.

        fopt : float, default 0.5
            Float number between 0 and 1, representing the fraction of
            remaining LORs in a sample used to locate the particle.

        verbose : bool, default False
            Print extra information when initialising this class.
        '''

        # Use @fopt.setter (below) to do the relevant type-checking when
        # setting fopt (self._fopt is the internal attribute, that we only
        # access through the getter and setter of the self.fopt property).
        self.fopt = fopt

        if verbose:
            print("Initialised BirminghamMethod.")


    @property
    def fopt(self):
        return self._fopt


    @fopt.setter
    def fopt(self, new_fopt):
        new_fopt = float(new_fopt)
        if new_fopt > 1 or new_fopt <= 0:
            raise ValueError(textwrap.fill(
                "[ERROR]: fopt should be set between 0 and 1. Received "
                f"{new_fopts}."
            ))

        self._fopt = new_fopt


    # Use the standardised function names `fit_sample` (for one numpy array)
    # and `fit` (for PointData or LineData).
    def fit_sample(
        self,
        sample,
        get_used = False,
        as_array = True,
        verbose = False
    ):
        '''Use the Birmingham method to track a tracer location from a numpy
        array (i.e. one sample) of LoRs.

        For the given `sample` of LoRs (a numpy.ndarray), this function
        minimises the distance between all of the LoRs, rejecting a fraction of
        lines that lie furthest away from the calculated distance. The process
        is repeated iteratively until a specified fraction (`fopt`) of the
        original subset of LORs remains.

        Parameters
        ----------
        sample : (N, M>=7) numpy.ndarray
            The sample of LORs that will be clustered. Each LoR is expressed as
            a timestamps and a line defined by two points; the data columns are
            then `[time, x1, y1, z1, x2, y2, z2, extra...]`.

        get_used : bool, default False
            If `True`, the function will also return a boolean mask of the LoRs
            used to compute the tracer location - that is, a vector of the same
            length as `sample`, containing 1 for the rows that were used, and 0
            otherwise.

        as_array : bool, default True
            If set to True, the tracked locations are returned as numpy arrays.
            If set to False, they are returned inside an instance of
            `pept.PointData` for ease of iteration and plotting.

        verbose : bool, default False
            Provide extra information when tracking a location: time the
            operation and show a progress bar.

        Returns
        -------
        locations : numpy.ndarray or pept.PointData
            The tracked locations found.

        used : numpy.ndarray, optional
            If `get_used` is true, then also return a boolean mask of the LoRs
            used to compute the tracer location - that is, a vector of the same
            length as `sample`, containing 1 for the rows that were used, and 0
            otherwise.
            [ Used for multi-particle tracking, not implemented yet]

        Raises
        ------
        ValueError
            If `sample` is not a numpy array of shape (N, M), where M >= 7.
        '''

        if verbose:
            start = time.time()

        # Type-check input parameters.
        # sample cols: [time, x1, y1, z1, x2, y2, z2, etc.]
        sample = np.asarray(sample, dtype = float, order = "C")

        if sample.ndim != 2 or sample.shape[1] < 7:
            raise ValueError(textwrap.fill(
                "[ERROR]: `sample` should have two dimensions (M, N), where "
                f"N >= 7. Received {sample.shape}."
            ))

        locations, used = birmingham_method(sample, self._fopt)

        if not as_array:
            locations = pept.PointData(
                locations, sample_size = 0, overlap = 0, verbose = False
            )

        if verbose:
            end = time.time()
            print((
                "Tracking one location with %i LORs took %.3f seconds" %
                (sample.shape[0], end - start)
            ))

        if get_used:
            return locations, used

        return locations


    # Use the standardised function names `fit_sample` (for one numpy array)
    # and `fit` (for PointData or LineData).
    def fit(
        self,
        line_data,
        max_error = 10,
        get_used = False,
        max_workers = None,
        verbose = True
    ):
        '''Fit lines of response (an instance of 'LineData') and return the
        tracked locations and (optionally) the LoRs that were used.

        This is a convenience function that asynchronously iterates through the
        samples in a `LineData`, finding the tracer locations. For more
        fine-grained control over the tracking, the `fit_sample` method can be
        used for individual samples.

        Parameters
        ----------
        line_data : an instance of `pept.LineData`
            The samples of lines of reponse (LoRs) that will be used for
            locating the tracer. Be careful to set the appropriate
            `sample_size` and `overlap` for good results. If the `sample_size`
            is too low, the tracer might not be found; if it is too high,
            temporal resolution is decreased. If the `overlap` is too small,
            the tracked points might be very "sparse".

        max_error : float, default = 10
            The maximum error allowed to return a 'valid' tracked location. All
            tracer locations with an error larger than `max_error` will be
            discarded.

        get_used : bool, default False
            If `True`, the function will also return a list of boolean masks of
            the LoRs used to compute the tracer location for each sample - that
            is, a vector of the same length as `sample`, containing 1 for the
            rows that were used, and 0 otherwise.

        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            clustering the samples in `cutpoints`. If unset (`None`), the
            number of threads available on the machine (as returned by
            `os.cpu_count()`) will be used.

        verbose : bool, default True
            Provide extra information when tracking: time the operation and
            show a progress bar.

        Returns
        -------
        locations : pept.PointData
            The tracer locations found.

        used : list of numpy.ndarray
            A list of boolean masks of the LoRs used to compute the tracer
            location for each corresponding sample in `line_data` - that is, a
            vector of the same length as a sample, containing 1 for the rows
            that were used, and 0 otherwise.

        Raises
        ------
        TypeError
            If `line_data` is not an instance of `pept.LineData`.
        '''

        if verbose:
            start = time.time()

        if not isinstance(line_data, pept.LineData):
            raise TypeError(textwrap.fill(
                "[ERROR]: `line_data` should be an instance of `pept.LineData`"
                f" (or any subclass thereof). Received {type(line_data)}."
            ))

        # Users might forget to set the sample_size, leaving it to the default
        # value of 0; in that case, all lines are returned as a single sample -
        # that might not be the intended behaviour.

        if line_data.sample_size == 0:
            warnings.warn(
                textwrap.fill((
                    "\n[WARNING]: The `line_data.sample_size` was left to the "
                    "default value of 0, in which case all lines are returned "
                    "as a single sample. For a very large number of lines, "
                    "this might result in a long function execution time.\n"
                    ), replace_whitespace=False),
                RuntimeWarning
            )

        get_used = bool(get_used)

        # Using ThreadPoolExecutor, asynchronously collect the locations from
        # every sample in a list of arrays. This is more efficient than using
        # ProcessPoolExecutor (or joblib) because birmingham_method is a Cython
        # function that releases the GIL for most of its computation.
        # If verbose, show progress bar using tqdm.
        if max_workers is None:
            max_workers = os.cpu_count()

        with ThreadPoolExecutor(max_workers = max_workers) as executor:
            futures = []
            for sample in line_data:
                futures.append(
                    executor.submit(
                        birmingham_method,
                        sample,
                        self._fopt
                    )
                )

            if verbose:
                futures = tqdm(futures)

            data_list = [f.result() for f in futures]

        # Access the data_list output as list comprehensions
        # data_list is a list of tuples, in which the first element is an
        # array of the `location`, and the second element is `used`, a
        # boolean mask representing the used LoRs.
        locations = [r[0] for r in data_list if len(r[0]) != 0]
        used = [r[1] for r in data_list if len(r[1]) != 0]

        # Remove LoRs with error above max_error
        locations = np.vstack(locations)
        locations = np.delete(
            locations,
            np.argwhere(locations[:, 4] > max_error),
            axis = 0
        )

        if len(locations) != 0:
            locations = pept.PointData(
                locations,
                sample_size = 0,
                overlap = 0,
                verbose = False
            )

        if verbose:
            end = time.time()
            print("\nTracking locations took {} seconds\n".format(end - start))

        if get_used:
            # `used is a list of the `used` arrays for the corresponding sample
            # in `line_data`.
            return locations, used

        return locations


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"fopt = {self.fopt}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL
        docstr = (
            "Class instance that inherits from `BirminghamMethod`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n"
            "----------\n"
            f"{self.__str__()}\n"
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.08.2019


'''Transform data from different PET / PEPT scanner geometries and data formats
into the common base classes.

Summary
-------
The PEPT base classes `PointData`, `LineData`, and `VoxelData` are abstractions
over the type of data that may be encountered in the context of PEPT (e.g. LoRs
are `LineData`, trajectory points are `PointData`), as once the raw data is
transformed into the common formats, any tracking, analysis or visualisation
algorithm in the `pept` package can be used interchangeably.

The `pept.scanners` subpackage provides modules for transforming the raw data
from different PET / PEPT scanner geometries (parallel screens, modular
cameras, etc.) and data formats (binary, ASCII, etc.) into the common base
classes.

If you'd like to integrate another scanner geometry or raw data format into
this package, you can check out the `pept.scanners.parallel_screens` module as
an example. This usually only involves writing a single function by hand; then
all attributes and methods from `LineData` will be available to your new data
format.

Subpackages Provided
--------------------

::

    pept.scanners
    │
    Classes imported into subpackage root:
    ├── ModularCamera :    Convert modular cameras data into `LineData`.
    ├── ParallelScreens :  Initialise parallel screens data as `LineData`.
    ├── ADACForte :        Initialise `LineData` from an ADAC binary file.
    │
    Subpackages
    ├── modular_camera
    │   └── ModularCamera
    └── parallel_screens
        ├── ParallelScreens
        └── ADACForte

'''


from    .parallel_screens    import  *
from	.modular_camera	     import  *


__author__ = ["Andrei Leonard Nicusan", "Sam Manger"]
__credits__ = [
    "Andrei Leonard Nicusan",
    "Kit Windows-Yule",
    "Sam Manger"
]
__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Development"
